#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
@File    :   main.py
@Time    :   2024/01/03 08:19:02
@Author  :   Vlad Ungureanu
@Version :   1.0
@Contact :   vlad.ungureanu@york.ac.uk
@Desc    :   This is the main file of the network pipeline developed to create a network from TPMs and run different community detection algorithms.

The scripts has the following capabilities:
1. Create a co-regulated network based on TPMs
2. Selective edge pruning can configured
3. Run different community detection algorithm
"""

import pandas as pd
import numpy as np
import os
import sys
import time
import plotly.express as px
import math as math
import igraph as ig
import leidenalg as la
import random
from scipy.stats import rankdata
import pickle as pickle
import graph_tool.all as gt


def measure_execution_time(func):
    """Decorator to measure the execution time of a function."""
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time
        hours = int(execution_time // 3600)
        minutes = int((execution_time % 3600) // 60)
        seconds = execution_time % 60
        print(f"Function {func.__name__} took {minutes}:{seconds:.2f} s to execute")
        return result

    return wrapper


class iCoExpNet:

    # Graph construction parameters
    edges_pg: int
    edges_sel: int
    genes_kept: int
    retain_f = 0.5  # 50%

    # Weight modifier parameter
    modifier_type = "standard"

    # Comm detection type
    graph_type = "gt"  # gt or ig - for leiden
    sbm_method = "hsbm"
    sbm_config: dict

    # Leiden algorithm parameter
    mod_type = "mod_max"  # or CPM (Constant Pots Model)
    resolution_parameter = None  # set only when the CPM mode is selected

    save_to_gephi = False

    # general
    g: ig.Graph # For igraph object - used for leiden algorithm
    gt_g: gt.Graph # For graph-tool object - with SBM & hSBM
    exp_name: str

    # other settings
    gene_sel_type = "rel_std"
    round_decimal = 3

    meta_exists: bool

    # folders:
    in_f: str
    out_f: str

    # Intermediate folders/files paths
    inpute_ge_file: str
    sel_ge_file: str
    mut_file: str
    master_strats: str
    output_folder: str
    prcsd_data_folder: str

    def __init__(
        self, exp_name, ge_file:str, input_folder:str, output_folder:str, gene_subset_file:str, mut_file:str, genes_kept=5000, edges_pg=3, edges_sel=6, modifier_type="standard", mod_type="mod_max", **kwargs
    ) -> None:
        """Initialize the iCoExpNet object.

        Args:
            exp_name (str): The name of the experiment.
            ge_file (str): The gene expression file.
            input_folder (str): The input folder path.
            output_folder (str): The output folder path.
            gene_subset_file (str): The file containing the subset of genes to be used for edge pruning.
            mut_file (str): The file containing the mutations data used for modifying the edges weights.
            genes_kept (int, optional): The number of genes to keep from the original dataset. Defaults to 5000.
            edges_pg (int, optional): The number of edges for standard genes. Defaults to 3. From PGCNA (Care. et al. 2020) lower values when using Leiden algorithm led to disconnected communities.
            edges_sel (int, optional): The number of edges for the subset of genes to have higher number of edges in the graph. Defaults to 6. Value 6 was chosen based on the experiments in Vlad Ungureanu's PhD thesis where higher values had diminishing returns.
            modifier_type (str, optional): The type of weight modifier to use. Defaults to "standard" which doesn't change the edges weights.
            mod_type (str, optional): The type of modularity optimization to use if Leiden algorithm is selected: "mod_max" or "CPM". Defaults to "mod_max".
        """

        self.exp_name = exp_name
        self.genes_kept = genes_kept
        self.edges_pg = edges_pg
        self.edges_sel = edges_sel
        self.modifier_type = modifier_type
        self.mod_type = mod_type


        # check if self.sel_ge_file exists
        self.sel_ge_file = os.path.abspath(f"{input_folder}/{gene_subset_file}")
        if not os.path.exists(self.sel_ge_file):
            raise FileExistsError(f"There is no file for Selected genes at {self.sel_ge_file}")

        # if we change the edges weights we need to have the mutation files 
  
        self.mut_file = os.path.abspath(f"{input_folder}/{mut_file}")
        # check if self.mut_file exists
        if not os.path.exists(self.mut_file):
            raise FileNotFoundError(f"There is no file with mutations at {self.mut_file}")

        if "graph_type" in kwargs.keys():
            self.graph_type = kwargs["graph_type"]

        if "sbm_method" in kwargs.keys():
            self.sbm_method = kwargs["sbm_method"]

        if "sbm_config" in kwargs.keys():
            self.sbm_config = kwargs["sbm_config"]
            self.sbm_method = kwargs["sbm_config"]["method"]
        else:
            self.sbm_config = {"n_iter": 1000, "mc_iter": 10, "deg_cor": True}

        if mod_type == "CPM":
            self.resolution_parameter = kwargs["res_param"]

        self.set_paths(exp_name, in_f=input_folder, ge_file=ge_file, out_f= output_folder)

    def compute_g_matrix(self):
        """
        Computes the graph matrix by loading data, preprocessing TPMs, computing the correlation matrix,
        applying modifiers, reducing edges, and exporting the result to Gephi.

        Returns:
            pruned_df (DataFrame): Pruned correlation matrix.
            corr_df (DataFrame): Correlation matrix.
            ge_df (DataFrame): Preprocessed TPMs.
            sel_ge (DataFrame): Selected gene expression data.
        """

        # still load the data
        ge_df, sel_ge, mut_df = self.load_data(ge_path=self.input_ge_file, sel_ge_path=self.sel_ge_file, mut_path=self.mut_file)

        #
        meta_corr_df, prcsd_meta_df = self.load_prcsd_meta()
        if meta_corr_df is None:

            # Pre-process TPMs
            ge_df = self.filter_data(ge_df, num_genes=self.genes_kept, type=self.gene_sel_type)

            # Compute corr matrix
            corr_df = self.corr_matrix(df=ge_df).round(self.round_decimal)

            # apply modifiers
            corr_df = self.weight_modifier(ge_df=ge_df, corr_df=corr_df, mut_df=mut_df, modifier_type=self.modifier_type)

            self.save_prcsd_data(corr_df, prcsd_meta_df)
        else:
            corr_df = meta_corr_df

        pruned_df = self.reduce_edges(corr_df=corr_df.copy(deep=True), ge_df=ge_df, sel_ge=sel_ge)

        # For debugging - intermediate files
        # pruned_df.to_csv(f'{self.master_stats}/pruned_df.tsv', sep="\t", index=True)
        # corr_df.to_csv(f'{self.master_stats}/modified_corr_df.tsv', sep="\t",  index=True)
        # ge_df.to_csv(f'{self.master_stats}/ge_df.tsv', sep="\t",  index=True)

        return pruned_df, corr_df, ge_df, sel_ge

    def set_paths(self, exp_name:str, in_f:str, ge_file:str, out_f:str):
        """
        Set the file paths for input and output files.

        Args:
            exp_name (str): The name of the experiment.
            in_f (str): The input file directory.
            out_f (str): The output file directory.
        """
        self.input_ge_file = f"{in_f}/{ge_file}"
        self.master_stats = f"{out_f}/Stats/"
        self.output_folder = f"{out_f}/Networks/{exp_name}/EPG{self.edges_pg}/"
        self.prcsd_data_folder = f"{out_f}/Processed/"

        print(f"#### Results are going to be saved in {self.output_folder}")

        if not os.path.exists(self.output_folder):
            os.makedirs(self.output_folder)

        if not os.path.exists(self.master_stats):
            os.makedirs(self.master_stats)

        if not os.path.exists(self.prcsd_data_folder):
            os.makedirs(self.prcsd_data_folder)

    def set_sel_ge(self, folder_path: str, filename: str): 
        """Load the files with the subset genes which are prioritised for edge pruning.

        Args:
            folder_path (str): The path to the folder containing the selected gene file.
            filename (str): The name of the selected gene file.
        """

    def save_experiment(self, sel_ge: list):
        """
        Save a snapshot of the experiment run.

        Parameters:
        - sel_ge (list): List of selected genes.

        Returns:
            None
        """

        save_dict = {
            "exp_name": self.exp_name,
            "graph_type": self.graph_type,
            "sbm_method": self.sbm_method,
            "modifier_type": self.modifier_type,
            "edges_pg": self.edges_pg,
            "edges_sel": self.edges_sel,
            "genes_kept": self.genes_kept,
            "retain_f": self.retain_f,
            "gene_sel_type": self.gene_sel_type,
            "resolution_parameter": self.resolution_parameter,
            "input_ge_file": self.input_ge_file,
            "input_mut_file": self.mut_file,
        }

        # this holds a history of the experiments run
        masterFile = "stats_master.tsv"
        masterDf = pd.DataFrame()
        masterFilePath = os.path.join(self.master_stats, masterFile)
        if os.path.exists(masterFilePath):
            masterDf = pd.read_csv(masterFilePath, sep="\t", engine="pyarrow")

        masterDf = pd.concat([masterDf, pd.DataFrame.from_dict(save_dict, orient="index").T])
        masterDf.to_csv(masterFilePath, sep="\t", index=False)

        save_dict["graph"] = self.g
        save_dict["tf_list"] = sel_ge
        filename = self.exp_name + ".pickle"
        objPath = os.path.join(self.master_stats, filename)
        with open(objPath, "wb") as handle:
            pickle.dump(save_dict, handle, protocol=pickle.HIGHEST_PROTOCOL)

    ############ Create graph ############
    @measure_execution_time
    def load_data(self, ge_path: str, sel_ge_path: str, mut_path: str):
        """
        Load data from a file.

        Args:
            ge_path (str): The path to the file.
            sel_ge_path (str): The path to selected genes in edge pruning
            mut_path (str): The path to TCGA mutations

        Returns:
            pandas.DataFrame: The loaded data as a DataFrame.

        Raises:
            FileExistsError: If the file does not exist.
        """
        # check if path exists
        if not os.path.exists(ge_path):
            raise FileExistsError(f"There is no file for TPMs at {ge_path}")

        if not os.path.exists(mut_path):
            raise FileExistsError(f"There is no file for TCGA mutations at {mut_path}")

        sel_ge = None
        if sel_ge_path is not None:
            if not os.path.exists(sel_ge_path):
                raise FileExistsError(f"There is no file for Selected gene at {sel_ge_path}")

            # TODO: change this to load the file with pandas
            sel_ge = np.genfromtxt(fname=sel_ge_path, delimiter="\t", skip_header=1, dtype="str")

        df = pd.read_csv(ge_path, index_col="gene", sep="\t", engine="pyarrow")
        
        df_mut = pd.read_csv(mut_path, index_col="gene", engine="pyarrow", sep="\t")
        return df, sel_ge, df_mut

    @measure_execution_time
    def corr_matrix(self, df: pd.DataFrame, method="spearman"):
        """
        Calculate the correlation matrix for a given DataFrame.

        Parameters:
            df (pd.DataFrame): The DataFrame for which the correlation matrix needs to be calculated.
            method (str, optional): The method used to calculate the correlation. Defaults to "spearman".

        Returns:
            pd.DataFrame: The correlation matrix.
        """

        if method == "partial_corr":
            corr_df = df.T.pcorr() # type: ignore it can be used with pingouin
        else:
            corr_df = df.T.corr(method=method) # type: ignore
        return corr_df


    @measure_execution_time
    def weight_modifier(self, ge_df: pd.DataFrame, corr_df: pd.DataFrame, mut_df: pd.DataFrame, modifier_type: str):
        """
        Modifies the weights of a correlation DataFrame based on a given modifier type.

        Args:
            corr_df (pd.DataFrame): The correlation DataFrame.
            mut_df (pd.DataFrame): The mutation DataFrame.
            modifier_type (str): The type of modifier to be applied.

        Returns:
            pd.DataFrame: The modified correlation DataFrame.
        """

        def rescaled(series: pd.Series, new_max: int):
            """Rescales a given series to a new range.

            Args:
                series (pd.Series): The series to be rescaled.
                new_max (int): The maximum value of the new range.

            Returns:
                pd.Series: The rescaled series.
            """
            new_min = -new_max
            x_norm = (series - series.min()) / (series.max() - series.min())
            x_scaled = x_norm * (new_max - new_min) + new_min

            return x_scaled

        def sigmoid_func(x, x0, offset=1) -> pd.DataFrame:
            """
            Sigmoid function

            Args:
                x (int/float): The variable
                x0 (center): Where the sigmoid is centred
                offset (int, optional): The elongation on y-axis. Defaults to 1.

                By default the sigmoid function starts goes from 1 to 2. The offset elongates on the y axis

            Returns:
                _type_: _description_
            """
            return (1 + math.exp(-(x - x0))) ** -1 * offset + 1

        if modifier_type == "standard":
            return corr_df

        # the first 2 lines ensures that all the mut_counts are found for the used genes
        ge_df["mut_count"] = mut_df['count']
        ge_df["mut_count"] = ge_df["mut_count"].fillna(0)
        
        mut = ge_df["mut_count"]
        mut_log = np.log2(mut + 1)
        max_log = mut_log.max()

        modifier = None
        if modifier_type == "reward":
            modifier = (max_log + mut_log) / max_log
        elif modifier_type == "penalised":
            modifier = (max_log - mut_log) / max_log
        elif modifier_type == "sigmoid":
            # 12, -12 was found through experimentations
            x0, offset, new_max = -8, 1, 12
            modifier = rescaled(mut, new_max=new_max)
            modifier = modifier.apply(sigmoid_func, args=(x0, offset))
        else:
            raise ValueError(f"Invalid modifier type: {modifier_type}")

        modifier.name = modifier_type
        modifier.replace(0, 1, inplace=True)

        # some adjustment
        corr_df = corr_df * modifier

        ge_df.drop(columns=['mut_count'], axis=1, inplace=True)

        return corr_df

    @measure_execution_time
    def create_graph(self, df: pd.DataFrame) -> ig.Graph:
        """
        Create a graph based on the given DataFrame.

        Parameters:
        df (pd.DataFrame): The DataFrame containing the graph data.

        Returns:
        g (ig.Graph): The created graph.
        edge_list (list of tuples): Needed to create graph tool
        """

        # remove self-loop
        dmy = df.replace(1, 0).values
        g = ig.Graph.Adjacency((dmy > 0).tolist(), mode="max")
        g.vs["name"] = df.index

        edges_list = []
        for edge in g.es:
            source_id, target_id = edge.tuple
            # dealing with 0
            source_target = df.loc[g.vs[source_id]["name"], g.vs[target_id]["name"]]
            target_source = df.loc[g.vs[target_id]["name"], g.vs[source_id]["name"]]
            weight = max(target_source, source_target)
            edge["weight"] = weight
            edges_list.append((g.vs[source_id]["name"], g.vs[target_id]["name"], weight))

        self.g = g
        return g, edges_list

    def create_gt(self, edges):
        """
        Creates the Graph Tool Graph based on the list of edges generated with PGCNA
        """
        edges_df = pd.DataFrame(edges, columns=["Source", "Target", "Weight"])

        list_of_edges = list(edges_df[["Source", "Target", "Weight"]].itertuples(index=False, name=None))
        gt_g = gt.Graph(directed=False)
        vertex_prop = gt_g.add_edge_list(edge_list=list_of_edges, hashed=True, eprops=[("weight", "double")])
        gt_g.vp.gene = vertex_prop

        self.gt_g = gt_g

        return gt_g

    @measure_execution_time
    def create_graph_from_gephi(self, gephi_path):
        # The role of this function is to create a graph from the Gephi file and then compared with the create_graph function.
        print("Work in progress")

    @measure_execution_time
    def export_to_gephi(self, pruned_df: pd.DataFrame, ge_df: pd.DataFrame, sel_ge: np.array):
        """
        Export the data to Gephi format.

        Args:
            pruned_df (pd.DataFrame): The pruned DataFrame containing the gene data.
            ge_df (pd.DataFrame): The DataFrame containing the gene expression data.
            sel_ge (np.array): The array of selected genes.

        Returns:
            None
        """

        iCoExpNet.print_msg("Exporting to Gephi")

        # create the nodes file
        nodes_df = pd.DataFrame(index=ge_df.index)
        nodes_df["Label"] = nodes_df.index
        nodes_df["TF"] = 0
        nodes_df.loc[nodes_df.index.isin(sel_ge), "TF"] = 1
        iCoExpNet.save_df(nodes_df, path=f"{self.output_folder}/Gephi/", filename="nodes.tsv")
        iCoExpNet.print_msg("Exported Nodes df")

        history = set()
        edges = []
        # create the edge file
        for gene_src, row in pruned_df.iterrows():
            # Select non-zero and different than itself
            sel_edges = row[(row != 0.0) & (row != 1.0)]

            new_items = []
            for gene_trg, weight in sel_edges.items():
                link_1 = gene_src + gene_trg
                link_2 = gene_trg + gene_src
                if (link_1 in history) or (link_2 in history):
                    continue

                edges.append((gene_src, weight, gene_trg))
                new_items.extend([link_1, link_2])

                history = history | set(new_items)

        edges = pd.DataFrame(edges, columns=["Source", "Weight", "Target"])
        edges["Type"] = "undirected"
        # do we need this?
        edges["fromAltName"] = edges["Source"]
        edges["toAltName"] = edges["Target"]

        iCoExpNet.save_df(edges, path=f"{self.output_folder}/Gephi/", filename="edges.tsv")
        iCoExpNet.print_msg("Exported Nodes df")
        iCoExpNet.print_msg("Gephi export completed")

    @measure_execution_time
    def compute_stats(self, ge_df: pd.DataFrame) -> pd.DataFrame:
        """
        Compute statistics for gene expression data from:

            -- quartile to expression (from percentiles)
            -- median (Q2) varWithin
            -- varAcross -  Quartile Coefficient of Dispersion (from percentiles)

        These values are used to later compute ModuleConnection and ModuleEvaluation.


        NOTE: This is an adaptation of _generateGeneMeta_ function from PGCNA code (Care et al.). It returns the same values
        Args:
            ge_df (pd.DataFrame): DataFrame containing gene expression data.

        Returns:
            pd.DataFrame: DataFrame containing computed statistics for each gene.
        """

        def listToPercentiles(x: np.array) -> np.array:
            data = rankdata(x, "average") / float(len(x))
            return data * 100

        def expToPercentiles(expToPerc: dict, expA: np.array) -> list:
            return [expToPerc[e] for e in expA]

        def dataSpread(x):
            """
            Returns the min, Q1 (25%), median (Q2), Q3 (75%), max, IQR, Quartile Coefficient of Dispersion and IQR/Median (CV like)

            Args:
                x (np.array): The input array.

            Returns:
                tuple: A tuple containing the min, Q1, median, Q3, max, IQR, QCOD, and IQR/Median.
            """

            q1 = float(np.percentile(x, 25, method="lower"))
            q2 = np.percentile(x, 50)
            q3 = float(np.percentile(x, 75, method="higher"))

            if (q2 == 0) or ((q3 + q1) == 0):
                return min(x), q1, q2, q3, max(x), abs(q3 - q1), 0, 0
            else:
                return min(x), q1, q2, q3, max(x), abs(q3 - q1), abs((q3 - q1) / (q3 + q1)), abs((q3 - q1) / q2)

        gene_stats = {}
        # Flatten the values to have all the expression
        expressionVals = ge_df.values.flatten()

        # Convert expression to percentiles
        expPercentiles = listToPercentiles(expressionVals)

        #  Calculate mapping from expression to percentile
        expToPerc = {}
        for i, e in enumerate(expressionVals):
            expToPerc[e] = expPercentiles[i]

        for i, gene in enumerate(ge_df.index):
            #  Convert from expression values to percentiles, so can work out (mean absolute deviation) MAD of percentiles
            genePercentiles = expToPercentiles(expToPerc, ge_df.values[i])

            try:
                minE, q1E, q2E, q3E, maxE, iqrE, qcodE, iqrME = dataSpread(ge_df.values[i])
            except:
                print("Issue with :", gene, ge_df.values[i])
                sys.exit()

            medianPercentiles = np.median(genePercentiles)
            # quartile to expression, median (Q2) varWithin, varAcross -  Quartile Coefficient of Dispersion
            gene_stats[gene] = [q2E, qcodE, medianPercentiles]

        gene_stats = pd.DataFrame.from_dict(gene_stats, orient="index", columns=["q2E", "qcodE", "median"])

        iCoExpNet.save_df(gene_stats, f"{self.output_folder}/", filename="gene_stats.tsv")
        return gene_stats

    ############ Community detection algorithms ############
    @measure_execution_time
    def run_leiden(self, g: ig.Graph, mod_type="mod_max", runs=100, res_param=None):
        """
        Runs the Leiden algorithm on a given graph.

        Args:
            g (ig.Graph): The input graph.
            mod_type (str, optional): The modularity type. Defaults to "mod_max".
            num (int, optional): The number of runs. Defaults to 100.
            res_param (float, optional): The resolution parameter. Defaults to None.

        Returns:
            pd.DataFrame, {}: A DataFrame containing the results of the Leiden algorithm and a dictionary containing the top 10 partitions

        """
        iCoExpNet.print_msg(f"Start Leiden with mod_type = {mod_type}; runs = {runs}; res_param={res_param}")
        print_count = int(runs / 10)

        graph_weights = g.es["weight"]
        scores, partitions = [], {}
        for run in range(1, runs + 1):
            seed = random.randint(0, int(1e9))  # Generate random seed as leidenalg doesn't appear to do this (even though it says it does)

            if mod_type == "mod_max":
                partition = la.find_partition(
                    g, la.ModularityVertexPartition, weights=graph_weights, n_iterations=-1, seed=seed
                )  # n_iterations=-1 to converge on best local result
            else:
                partition = la.find_partition(g, la.CPMVertexPartition, weights=graph_weights, n_iterations=-1, seed=seed, resolution_parameter=res_param)

            part_sizes = partition.sizes()
            partitions[run] = partition
            scores.append(
                (
                    partition.modularity,
                    sum(part_sizes) / len(part_sizes),
                    len(part_sizes),
                    part_sizes,
                )
            )

            if run % print_count == 0:
                print(f"### Run {run}. ModularityScore={scores[-1][0]}; AvgModSize={scores[-1][1]}; ModuleNum={scores[-1][2]}")

        # Process results
        scores_df = pd.DataFrame(scores, columns=["ModularityScore", "AvgModSize", "ModuleNum", "ModuleSizes"])
        scores_df = scores_df.sort_values(by="ModularityScore", ascending=False).reset_index(drop=False, names="Run").reset_index(names="Mod#")
        scores_df["Mod#"] = scores_df["Mod#"] + 1

        # save top 10
        best_leidens = {}
        for _, row in scores_df.iterrows():
            sel_part = partitions[row["Run"] + 1]
            names = sel_part.graph.vs["name"]
            members = sel_part.membership

            partition_df = pd.DataFrame(zip(names, members), columns=["Gene", "Modularity Class"])

            iCoExpNet.save_df(partition_df.set_index("Gene"), f"{self.output_folder}Leiden/Best/", filename=f"leiden_best_{row['Mod#']}.tsv")

            best_leidens[row["Mod#"]] = sel_part
            if row["Mod#"] == 10:
                break

        iCoExpNet.save_df(scores_df, f"{self.output_folder}Leiden/", filename="summary_leiden.tsv")
        iCoExpNet.print_msg(f"Finished Leiden")

        return scores_df, best_leidens

    @measure_execution_time
    def run_sbm(self, gt_g: gt.Graph, n_iter=10000, mc_iter=10, deg_cor=True, distrib="real-exponential", verbose=True):
        """
        Run the Stochastic Block Model (SBM) on a given graph.

        Args:
            gt_g (gt.Graph): The input graph.
            n_iter (int): The number of iterations for MCMC.
            mc_iter (int): The number of iterations for MCMC equilibration.
            deg_cor (bool): Whether to apply degree correction.
            distrib (str): The distribution type for the edge weights.
            verbose (bool): Whether to print verbose output.

        Returns:
            dict: The state object containing the partitions and other metadata.
        """

        if not hasattr(self, "states"):
            self.states = []
    
        state = gt.BlockState(gt_g, recs=[gt_g.ep.weight], rec_types=[distrib], deg_corr=deg_cor)

        print(f"### Equilibrate MCMC for {self.exp_name}")
        gt.mcmc_equilibrate(state, wait=n_iter / 10, mcmc_args=dict(niter=mc_iter), verbose=verbose)

        bs = []  # collect some partitions
        h = np.zeros(gt_g.num_vertices() + 1)  # to find the probabilities for a certain size
        dls = []  # description length history

        def collect_partitions(s):
            bs.append(s.b.a.copy())
            B = s.get_nonempty_B()
            h[B] += 1
            dls.append(s.entropy())

        print(f"### Collection partitions for {self.exp_name}")

        # Now we collect partitions for exactly 1000 sweeps, at intervals
        eq_res = gt.mcmc_equilibrate(state, force_niter=n_iter, mcmc_args=dict(niter=mc_iter), callback=collect_partitions, verbose=verbose)

        print(f"### Marginal property partitions for {self.exp_name}")
        # Disambiguate partitions and obtain marginals
        pmode = gt.PartitionModeState(bs, converge=True, relabel=True)
        print(f"### get all marginal for {self.exp_name}")
        pv = pmode.get_marginal(gt_g)

        new_state = {
            "name": f"state_{len(self.states) + 1}",
            "partition_meta": {"h": h, "n_iter": n_iter, "bs": bs},
            "dls": dls,
            "eq_res": eq_res,
            "pmode": pmode,
            "state": state,
            "pv": pv,
        }

        return new_state

    def run_hsbm(self, gt_g: gt.Graph, n_iter=10000, mc_iter=10, deg_cor=False):
        """
        Run the Hierarchical Stochastic Block Model (HSBM) on a given graph.

        Args:
            gt_g (gt.Graph): The input graph.
            n_iter (int): The number of iterations for MCMC.
            mc_iter (int): The number of iterations for MCMC equilibration.
            deg_cor (bool): Whether to apply degree correction.

        Returns:
            dict: The state object containing the partitions and other metadata.
        """
        state = gt.NestedBlockState(gt_g, recs=[gt_g.ep.weight], rec_types=["real-exponential"], state_args=dict(deg_corr=deg_cor))

        print(f"### 1. Equilibrate MCMC. n_iter {n_iter / 10}")
        # We will first equilibrate the Markov chain
        gt.mcmc_equilibrate(state, wait=n_iter / 10, mcmc_args=dict(niter=mc_iter), verbose=True)  # 1000 / 10 standard

        bs = []  # collect some partitions
        h = [np.zeros(gt_g.num_vertices() + 1) for s in state.get_levels()]
        dls = [np.zeros(gt_g.num_vertices() + 1) for s in state.get_levels()]

        # to find the probabilities for a certain size
        def collect_partitions(s):
            for l, sl in enumerate(s.get_levels()):
                B = sl.get_nonempty_B()
                h[l][B] += 1
            dls.append(s.entropy())
            bs.append(s.get_bs())

        print(f"### 2. Collection partitions. n_iter {n_iter}")
        eq_res = gt.mcmc_equilibrate(state, force_niter=n_iter, mcmc_args=dict(niter=mc_iter), callback=collect_partitions, verbose=True)  # 10000 / 10

        print("### 3. Marginal property partitions")
        # Disambiguate partitions and obtain marginals
        # pmode = gt.PartitionModeState(bs, nested=True, converge=False)
        # print("### 4. Collecting the data")
        # pv = pmode.get_marginal(gt_g)

        # Get consensus estimate
        # bs = pmode.get_max_nested()
        # state = state.copy(bs=bs)

        h_state_obj = {
            "name": "hstate",
            "partition_meta": {"h": h, "n_iter": n_iter, "bs": bs},
            "state": state,
            # "pv": pv,
            "dls": dls,
            "eq_res": eq_res,
        }

        return h_state_obj

    def dissambigue_hsbm_part(self, h_state_obj: dict):
        """
        Disambiguate the partitions from the Hierarchical Stochastic Block Model (hSBM) and obtain marginals.
        
        Args:
            h_state_obj (dict): The state object containing the partitions and other metadata.
        Returns:
            dict: The updated state object with disambiguated partitions and marginals.
        """
        print("### 3. Marginal property partitions")
        # Disambiguate partitions and obtain marginals
        pmode = gt.PartitionModeState(h_state_obj["partition_meta"]["bs"], nested=True, converge=False)
        print("### 4. Collecting the data")
        pv = pmode.get_marginal(self.gt_g)

        # Get consensus estimate
        bs = pmode.get_max_nested()
        state = h_state_obj["state"].copy(bs=bs)

        h_state_obj["partition_meta"]["bs"] = bs
        h_state_obj["pmode"] = pmode
        h_state_obj["state"] = state
        h_state_obj["pv"] = pv

        return h_state_obj

    ############ Utilities ############
    def sort_alphabetically(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Utility function to sort a DataFrame alphabetically by index and columns.
        
        Args:
            df (pd.DataFrame): The DataFrame to be sorted.
        Returns:
            pd.DataFrame: The sorted DataFrame."""
        df.sort_index(inplace=True)
        df = df[df.columns.sort_values()]
        return df

    @measure_execution_time
    def filter_data(self, ge_df: pd.DataFrame, num_genes=5000, type="rel_std"):
        """
        Filter the gene expression data based on a specified criterion.

        The standard is rel_std, but it also supports PGCNA style which is sorting by the median

        Parameters:
        ge_df (pd.DataFrame): The gene expression data.
        num_genes (int): The number of top genes to select.
        type (str): The type of criterion to use for filtering.

        Returns:
        pd.DataFrame: The filtered gene expression data.
        """
        if type == "rel_std":
            # so it doesn't divide by 0
            std_med = ge_df.std(axis=1) / ge_df.median(axis=1) # this is how it was run the sbm non_tum v3 (chapter 3 thesis)
            # std_med = ge_df.std(axis=1) / ge_df.mean(axis=1)

            top_genes = std_med.sort_values(ascending=False).index.values[:num_genes]
        elif type == "pgcna":
            std = ge_df.std(axis=1)
            top_genes = std.sort_values(ascending=False).index.values[:num_genes]
        else:
            top_genes = ge_df.index

        self.retain_f = round(len(top_genes) / ge_df.shape[0], 2)
        return ge_df.loc[ge_df.index.isin(top_genes)]

    @staticmethod
    def save_df(df: pd.DataFrame, path: str, filename: str):
        "Utility function to save a DataFrame to a file in TSV format."
        full_path = os.path.join(path, filename)

        # create directories if needed
        if not os.path.exists(path):
            os.makedirs(path)

        df.to_csv(full_path, sep="\t", index=True)

    @staticmethod
    def print_msg(message: str):
        "Utility function to print a message with a specific format."
        print(f"####### {message} #######")

    def save_obj(obj: any, path: str, filename: str):
        """Utility function to save an object to a file using pickle.
        Args:
            obj (any): The object to be saved.
            path (str): The directory path where the file will be saved.
            filename (str): The name of the file to save the object in.
        """        
        full_path = os.path.join(path, filename)

        # create directories if needed
        if not os.path.exists(path):
            os.makedirs(path)

        with open(full_path, "wb") as handle:
            pickle.dump(obj, handle, protocol=pickle.HIGHEST_PROTOCOL)

    ############ Intermediary files ############
    def load_prcsd_meta(self):
        """Utility function to load processed metadata from a .tsv file."""
        master_path = f"{self.prcsd_data_folder}/master.tsv"
        if os.path.exists(master_path):

            master_path = f"{self.prcsd_data_folder}/master.tsv"

            # dropping the empty index from the load
            prcsd_meta_df = pd.read_csv(master_path, sep="\t", engine="pyarrow").drop(columns=[""])

            sel_exp = prcsd_meta_df[prcsd_meta_df["num_genes"] == str(self.genes_kept)]
            if sel_exp.shape[0] == 0:
                return None, prcsd_meta_df

            corr_df = pd.read_csv(sel_exp["path"].values[0], sep="\t", engine="pyarrow", index_col="gene")
            return corr_df, prcsd_meta_df

        return None, pd.DataFrame()

    def save_prcsd_data(self, weighted_df: pd.DataFrame, prcsd_meta_df: pd.DataFrame):
        """
        Save processed data to a file and update the metadata. This avoids re-computing the same correlation matrix several times.
        Args:
            weighted_df (pd.DataFrame): The DataFrame containing the processed data.
            prcsd_meta_df (pd.DataFrame): The DataFrame containing the metadata.
        Returns:   
            None
        """
        if prcsd_meta_df is None:
            prcsd_meta_df = pd.DataFrame()

        corr_filename = f"corr_{self.modifier_type}.tsv"
        path_to_modifier = f"{self.prcsd_data_folder}/genes_{self.genes_kept}/"

        exp_meta = {
            "name": self.exp_name,
            "num_genes": self.genes_kept,
            "modifier": self.modifier_type,
            "edges_pg": self.edges_pg,
            "edges_sel": self.edges_sel,
            "path": f"{path_to_modifier}/{corr_filename}",
        }
        dmy_meta = pd.DataFrame().from_dict(exp_meta, orient="index").T
        if prcsd_meta_df.shape[0] == 0:
            prcsd_meta_df = dmy_meta
        else:
            prcsd_meta_df = pd.concat([prcsd_meta_df, dmy_meta], axis=0)

        iCoExpNet.save_df(weighted_df, path=path_to_modifier, filename=corr_filename)
        iCoExpNet.save_df(prcsd_meta_df.reset_index(drop=True), path=self.prcsd_data_folder, filename="master.tsv")

    def prcsd_meta_exists(self) -> bool:
        """
        Check if the processed data metadata file exists.

        Returns:
            bool: True if the metadata file exists, False otherwise.
        """
        master_path = f"{self.prcsd_data_folder}/master.tsv"
        if os.path.exists(master_path):
            prcsd_meta_df = pd.read_csv(master_path, sep="\t", engine="pyarrow")
            if prcsd_meta_df[prcsd_meta_df["name"] == self.exp_name].shape[0] != 0:
                print("Meta info exists!")
                return True
        else:
            return False

    ##### Testing
    @measure_execution_time
    @staticmethod
    def test_unique_edges(edges: pd.DataFrame):
        for gene in edges["Source"].unique():
            src_df = set(edges[edges["Source"] == gene]["Target"])
            trg_df = set(edges[edges["Target"] == gene]["Source"])

            if src_df & trg_df:
                print(f"Duplicated links for {gene}")
                print(f"Source {src_df}; Target {trg_df}")

    @staticmethod
    def make_zero(row, num_elems=3):
        """A utility function to set the bottom n elements of a row to zero.
        Args:
            row (pd.Series): The row to process.
            num_elems (int): The number of elements to set to zero.
        Returns:
            pd.Series: The processed row with the bottom n elements set to zero.
        """
        dmy_df = row.sort_values(ascending=False)
        dmy_df[num_elems + 1 :] = 0
        return dmy_df

    @staticmethod
    def bottomToZeroWithDuplicates(npA, n_dict: dict):
        """
        Set everything below n to zero,
        but deal with duplicates
        Args:
            npA (np.array): The numpy array to process.
            n_dict (dict): A dictionary containing the number of edges for each gene.
        Returns:
            np.array: The processed numpy array with values below n set to zero, while preserving duplicates.
        """

        num_e = n_dict[npA.name]
        unique = np.unique(npA)

        uniqueGTzero = len(unique[unique > 0])
        if num_e > uniqueGTzero:
            #  Deal with edgePG extending into negative correlations
            num_e = uniqueGTzero

        topIunique = np.argpartition(unique, -num_e)[-num_e:]
        toKeep = []
        for val in unique[topIunique]:
            # this ensures that the duplictates are kept too
            toKeep.extend(np.where(npA == val)[0])

        #  Mask and reverse
        mask = np.ones(len(npA), bool)
        mask[toKeep] = 0
        npA[mask] = 0

        return npA

    @measure_execution_time
    def reduce_edges(self, corr_df: pd.DataFrame, ge_df: pd.DataFrame, sel_ge: list):
        """The selective edge pruning function where we reduce the number of edges based on a subset genes.
        
        Args:
            corr_df (pd.DataFrame): The correlation DataFrame.
            ge_df (pd.DataFrame): The gene expression DataFrame.
            sel_ge (list): The list of selected genes.
        Returns:
            pd.DataFrame: The reduced correlation DataFrame."""
        # Compute the number of edges
        num_e_dict = pd.Series([self.edges_pg + 1] * ge_df.shape[0], index=ge_df.index, name="number_edges")
        num_e_dict.loc[num_e_dict.index.isin(sel_ge)] = self.edges_sel + 1
        num_e_dict = num_e_dict.to_dict()

        dmy_df = corr_df.apply(iCoExpNet.bottomToZeroWithDuplicates, args=(num_e_dict,), axis=1)

        return dmy_df

    @measure_execution_time
    def run(self):
        """Run the main analysis pipeline.

        Returns:
            None
        """
        # Dev
        def find_node(g: gt.Graph, gene_name='AHR'):
            for v in self.g.vertices():
                if self.g.vp['gene'][v] ==gene_name:
                    return v
                
        pruned_df, _, ge_df, sel_ge = self.compute_g_matrix()

        # stats df
        _ = self.compute_stats(ge_df)
        g, edges_list = self.create_graph(pruned_df)

        # Dev
        if 0:
            graph_stats = pd.DataFrame(index=g.vs[:]["name"])
            graph_stats["degree"] = g.degree()
            print(f"\n\n############ AHR iGRAPH degree {graph_stats.loc['AHR']['degree']}")  

            g2, edge_list = self.create_graph(pruned_df.T)
            graph_stats = pd.DataFrame(index=g.vs[:]["name"])
            graph_stats["degree"] = g.degree()
            # delete the large files
            # del pruned_df, ge_df

            graph_stats = pd.DataFrame(index=g.vs[:]["name"])
            graph_stats["degree"] = g2.degree()
            print(f"\n\n############ AHR iGRAPH 2 degree {graph_stats.loc['AHR']['degree']}")  


        # Running Leiden for comparison
        iCoExpNet.print_msg(f"Running Leiden")
        _, _ = self.run_leiden(g=g, mod_type="mod_max", runs=100)

        if self.graph_type == "gt":
            self.g = self.create_gt(edges_list)

            # Dev
            if 0:
                ahr_degree = self.g.get_total_degrees([find_node(self.g, gene_name='AHR')])
                print(f"\n\n############ AHR gt degree {ahr_degree}")   

            if self.sbm_method == "hsbm":
                iCoExpNet.print_msg(f"Running hSBM with {self.sbm_method}")
                state = self.run_hsbm(self.g, n_iter=self.sbm_config["n_iter"], mc_iter=self.sbm_config["mc_iter"], deg_cor=self.sbm_config["deg_cor"])

                # Sometimes disambiguity of the partitions it is failing
                iCoExpNet.save_obj(state, path=self.output_folder, filename=f"gt_{self.sbm_method}_{self.exp_name}.pickle")

                state = self.dissambigue_hsbm_part(state)
                
                # Update the graph object with the latest, after mcmc sweep
                self.gt_g = state['state'].g

            else:
                iCoExpNet.print_msg(f"Running SBM with {self.sbm_method}")
                state = self.run_sbm(self.g, n_iter=self.sbm_config["n_iter"], mc_iter=self.sbm_config["mc_iter"], deg_cor=self.sbm_config["deg_cor"])

                # Update the graph object with the latest, after mcmc sweep
                self.gt_g = state['state'].g

            iCoExpNet.save_obj(state, path=self.output_folder, filename=f"gt_{self.sbm_method}_{self.exp_name}.pickle")

        self.save_experiment(sel_ge=sel_ge)

        if self.save_to_gephi:
            self.export_to_gephi(pruned_df=pruned_df, ge_df=ge_df, sel_ge=sel_ge)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
from scipy.stats import spearmanr

from .utilities import sankey_consensus_plot as sky
from .utilities.helpers import save_fig
from .ExperimentSet import ExperimentSet
from .NetworkOutput import NetworkOutput


class NetworkComp:

    source_exp: NetworkOutput
    target_exp: NetworkOutput

    source: str
    target: str
    name: str

    comp_df: pd.DataFrame

    def __init__(self, exp_set: ExperimentSet, no_genes, source, target, mod_type = 'Leiden'):
        exp_labels = exp_set.get_exp_labels()

        # backwards compatibility
        if mod_type == "Leiden":
            if "_".join(source.split("_")[-3:]) not in exp_labels:
                raise ValueError("Source exp {} not in experiments {}".format(source, exp_set.name))

            if "_".join(target.split("_")[-3:]) not in exp_labels:
                raise ValueError("Target exp {} not in experiments {}".format(source, exp_set.name))

        # This may fail with non-sbm exps if so un-comment the next lines
        egTF = int(source.split("TF")[0].split("_")[-1])
        # if len(source.split("_")[-1]) > 3:
        #     egTF = int(source.split("_")[-1][:2])
        # else:
        #     egTF = int(source.split("_")[-1][:1])

        remap_nodes_df, _ = exp_set.run_comb_exp(no_genes=no_genes, egTF=egTF, comps=[(source, target)], mod_type=mod_type)
        remap_nodes_df.rename(columns={"count": "mut_count"}, inplace=True)

        # cols in nodes_df
        if "sbm" in mod_type: 
            source_col = f"max_b_{source}"
            target_col = f"max_b_{target}"
        else:
            source_col = f"Leiden_{source}"
            target_col = f"Leiden_{target}"

        # pre-process the data
        remap_df = remap_nodes_df.loc[:, ["mut_count", source_col, target_col]]
        remap_df["Source"] = [val.split("_")[-1] if "NA" not in val else np.nan for val in remap_df[source_col].values]
        remap_df["Target"] = [val.split("_")[-1] if "NA" not in val else np.nan for val in remap_df[target_col].values]

        # fill the NA with 0
        remap_df.loc[remap_df["mut_count"] == "NA", "mut_count"] = 0

        remap_df = remap_df[~remap_df["Source"].isnull()]

        self.name = "{} vs. {}".format(source, target)
        self.no_genes = no_genes
        self.comp_df = remap_df.copy(deep=True)
        # the above 2 may need renaming
        self.source = source_col
        self.target = target_col

        self.source_exp = exp_set.exps[source]
        self.target_exp = exp_set.exps[target]
        # NOTE: The above assignment work for SBM Experiments and it should work for Leiden as well, but un-comment the below code if it doesn't
        # self.source_exp = exp_set.exps[("_".join(source.split("_")[-3:]))]
        # self.target_exp = exp_set.exps[("_".join(target.split("_")[-3:]))]

    def compute_comm_mut_stats(self):
        dmy = {}

        if self.comm_stats == None:
            df = self.comp_df.copy(deep=True)
            for comm, orig_num in df["Source"].value_counts().items():
                stats = []
                diff = abs(orig_num - df[df["Target"] == comm].shape[0])

                sel_mut_orig = df[df["Source"] == comm]["mut_count"]
                sel_mut_changes = df[(df["Source"] == comm) & (df["Target"] != comm)]["mut_count"]

                # orig, changes, ratio
                stats.append(orig_num)
                stats.append(diff)
                stats.append(round(diff / orig_num, 4))

                # Mut stats
                stats.append(sel_mut_orig.mean())
                stats.append(sel_mut_changes.mean())

                stats.append(sel_mut_orig.median())
                stats.append(sel_mut_changes.median())

                dmy[comm] = stats

            comm_stats = (
                pd.DataFrame.from_dict(
                    data=dmy,
                    orient="index",
                    columns=[
                        "Orig_num",
                        "Changes_num",
                        "Ratio",
                        "Mut_orig_mean",
                        "Mut_changes_mean",
                        "Mut_orig_median",
                        "Mut_changes_median",
                    ],
                )
                .reset_index()
                .rename(columns={"index": "comm"})
            )
            self.comm_stats = comm_stats

    def com_mut_distrib(self, include_source=False, toSave=False, path=None, binarySource=False, annotations=None, ann_add=[], ann_rm=[], ann_chg=[]):
        if not include_source:
            df = self.comp_df[self.comp_df["Target"] != self.comp_df["Source"]].copy(deep=True)
        else:
            df = self.comp_df.copy(deep=True)

        title = "{}. Gene comm changes and mut counts".format(self.name)
        colors = px.colors.qualitative.Plotly + px.colors.qualitative.Safe

        # This enables to show on the histogram if the data came from other communities (any) or itself
        color_map = None
        if binarySource:
            df.loc[df["Target"] != df["Source"], "Source"] = 0
            # assign blue to incoming genes
            color_map = {0: colors[0]}
            colors.pop(0)

        order = sorted(df["Source"].astype(int).unique())
        fig = px.strip(
            df.reset_index(),
            x="Target",
            y="mut_count",
            color="Source",
            height=600,
            category_orders={"Target": order},
            title=title,
            color_discrete_sequence=colors,
            color_discrete_map=color_map,
            hover_data=["Id", "mut_count", "Source", "Target"],
        )
        fig.update_layout(
            legend=dict(
                orientation="v",
                title="Source",
                yanchor="middle",
                y=0.70,
                xanchor="center",
                x=0.96,
                bgcolor="rgba(0,0,0,0)",
            ),
        )

        if annotations:
            marker_genes, num_genes = [], 20
            marker_genes.extend(df.sort_values(by=["mut_count"], ascending=False).index.values[:num_genes])

            if len(ann_add) > 0:
                marker_genes.extend(ann_add)
            if len(ann_rm) > 0:
                marker_genes = set(marker_genes) - set(ann_rm)

            counter = 0
            offset = 10
            for idx, row in df.loc[df.index.isin(marker_genes)][["Target", "Source", "mut_count"]].iterrows():
                align_left = True
                if counter % 2 == 0:
                    shift = offset
                    align_left = False
                else:
                    shift = -offset

                if idx in ann_chg:
                    if align_left:
                        shift = -offset * 5.5
                    else:
                        shift = +offset * 5.5

                align = "left" if align_left else "right"
                x = str(int(row["Target"]) - 1)

                if row.name in ["CDKN2A", "FAT2"]:
                    align = "left"
                    shift = -offset * 7
                if row.name in ["COL7A1"]:
                    shift = offset

                fig.add_annotation(x=x, y=row["mut_count"], text=idx, showarrow=False, valign="top", align=align, xshift=shift, yshift=0)

        if toSave and path:
            fig_name = "box_{}_{}".format(self.source_exp.type, self.target_exp.type)
            save_fig(name=fig_name, fig=fig, base_path=path)
        else:
            return fig

    def membership_change(self, include_source=False, toSave=False, path=None):
        if not include_source:
            df = self.comp_df[self.comp_df["Target"] != self.comp_df["Source"]].copy(deep=True)
        else:
            df = self.comp_df.copy(deep=True)

        title = "{}. Gene comm changes".format(self.name)

        order = sorted(df["Source"].unique())

        fig = px.histogram(
            df,
            x="Target",
            color="Source",
            height=600,
            category_orders={"Target": order},
            title=title,
        )
        # fig.update_layout(
        #     legend=dict(
        #         orientation="v",
        #         title="Source",
        #         yanchor="middle",
        #         y=0.75,
        #         xanchor="center",
        #         x=0.96,
        #         bgcolor="rgba(0,0,0,0)",
        #     ),
        # )
        if toSave and path:
            fig_name = "memberShip_{}_{}".format(self.source_exp.type, self.target_exp.type)
            save_fig(name=fig_name, fig=fig, base_path=path)
        else:
            return fig

    def plt_median_ge(self, map_names: dict, path = None):

        comp_dict = self.comp_ge_comm()

        figs = []
        for key, df in comp_dict.items():
            df["Comm_num"] = df["Comm"].str.split("_", expand=True)[1]
            title = "{}. Median Gene Expression across communities.".format(map_names[key])
            fig_name = "P0_{}_med".format(key)
            fig = px.box(df, x="Comm_num", y="Median", color="Comm_num", title=title, points="all")
            fig.update_layout(showlegend=False, xaxis_title="Community", yaxis_title="Median Gene Expression")
            fig.update_traces(boxmean=True)
            figs.append(fig)

            if path != None:
                fig.update_layout(
                    title = "",
                    xaxis=dict(tickfont=dict(size=14)),
                    yaxis=dict(tickfont=dict(size=14), title="Median TPM"),
                    font=dict(size=16),
                )
                save_fig(name=fig_name, fig=fig, base_path=path, width=1200, height=400, margin=0.02)

        return figs

    def sankey_plot(self, df=None, drop_na=True, toSave=False, path=None, source_label=None, target_label=None):
        reoder_cols = [format(self.source), format(self.target)]
        if df is None:
            df = self.comp_df.copy(deep=True)

        # remove the genes added by 7K exp
        if drop_na:
            sel_idxs = df[reoder_cols][df[format(self.source)] != "NA"].dropna().index
            df = df.loc[sel_idxs]


        # This
        if (source_label != None) and (target_label != None):

            # Rename the rows values
            row_vals = self.source.replace("Leiden_", "")
            df[self.source] = df[self.source].str.replace(row_vals, source_label)

            row_vals = self.target.replace("Leiden_", "")
            df[self.target] = df[self.target].str.replace(row_vals, target_label)

            # Rename the columns
            df = df.rename(columns={self.target: target_label, self.source: source_label})
            reoder_cols = [source_label, target_label]

        if toSave and path:
            _, fig = sky.main(df, reorder_cols=reoder_cols, title="Community comparison between " + ", ".join(reoder_cols), retMeta=True)
            fig_name =f"sankey_{self.source_exp.type}_{self.target_exp.type}"
            save_fig(name=fig_name, fig=fig, base_path=path)
        else:

            _, fig = sky.main(df, reorder_cols=reoder_cols, title="Community comparison between " + ", ".join(reoder_cols), retMeta=True)
            return fig

    def find_mut_stats(self, th=1, col="Source"):
        dmy = []
        df = self.comp_df.copy(deep=True)
        for comm in df[col].unique():
            sel_comm = df[df[col] == comm]
            mut_genes = sel_comm[sel_comm["mut_count"] >= th].shape[0]
            if sel_comm.shape[0] != 0:
                percent = round(mut_genes / sel_comm.shape[0] * 100, 3)
            else:
                percent = 0

            dmy.append([comm, mut_genes, sel_comm.shape[0], percent])
        return pd.DataFrame(dmy, columns=[col, "Mut_genes", "All", "Percent"]).sort_values(by="Percent", ascending=False).set_index(col)

    def comb_mut_stats(self, start=0, end=60, direction="Source", log10=False):
        ths = list(range(start, end + 1, 1))

        all_dfs = pd.DataFrame()
        for th in ths:
            new_col = "{}".format(th)
            mut_target = self.find_mut_stats(th=th, col=direction).rename(columns={"Mut_genes": new_col})
            if th == start:
                all_dfs = mut_target[["All", new_col]].copy(deep=True)
            else:
                all_dfs = pd.concat([all_dfs, mut_target[new_col]], axis=1)

        dmy_df = all_dfs.melt(ignore_index=False, var_name="Mut_th", value_name="# genes").reset_index()
        if log10:
            dmy_df["# genes"] = np.log10(dmy_df["# genes"] + 1)
        return dmy_df, all_dfs

    def plot_mut_evo(self, df, direction="Source"):
        title = "{} comm. Mutations count in each community as we increase Mut_count threshold".format(direction)
        fig = px.line(
            df,
            x="Mut_th",
            y="# genes",
            color=direction,
            title=title,
            markers=True,
            color_discrete_sequence=px.colors.qualitative.G10)
        fig.update_layout(
            legend=dict(
                orientation="h",
                title=direction,
                yanchor="middle",
                y=0.95,
                xanchor="center",
                x=0.6,
                bgcolor="rgba(0,0,0,0)",
            ),
        )
        return fig

    def comp_ge_comm(self):
        ret_dict = {}
        source_mod = "_".join(self.source.split("_")[-3:])
        target_mod = "_".join(self.target.split("_")[-3:])

        ret_dict[source_mod] = self.source_exp.ge_comm()
        ret_dict[target_mod] = self.target_exp.ge_comm()

        return ret_dict

    def diff_in_com(self):
        remap_cols = {}
        for src_com in self.comp_df["Source"].unique():
            trgt_com = self.comp_df[self.comp_df["Source"] == src_com]["Target"].values[0]
            remap_cols[src_com] = trgt_com

        src_cols = [int(val) for val in remap_cols.values() if val is not np.nan]
        src_cols.sort()
        diff = set(self.target_exp.modCons.keys()) - set(src_cols)
        print("{} has the following communities in addition to {}: {}".format(self.target_exp.type, self.source_exp.type, diff))

    def plot_network_stats(self, nodes_df_1: pd.DataFrame, nodes_df_2: pd.DataFrame, label_1: str, label_2: str, path = None):

        # Add the stats
        graph_stats_1: pd.DataFrame = self.source_exp.compute_graph_stats()
        graph_stats_2: pd.DataFrame = self.target_exp.compute_graph_stats()

        graph_stats_1["mut_count"] = nodes_df_1["count"]
        graph_stats_2["mut_count"] = nodes_df_2["count"]

        graph_stats_1["TF"] = nodes_df_1["TF"]
        graph_stats_2["TF"] = nodes_df_2["TF"]

        # graph_stats_1["IVI"] = nodes_df_1["IVI"]
        # graph_stats_2["IVI"] = nodes_df_2["IVI"]

        # Define the metrics
        metrics = [
            # "degree_o",
            "degree_t",
            # "degree_w",
            "betweenness",
            "closeness",
            "katz",
            "pageRank",
            # "IVI",
            'mut_count'
            # "TF",
        ]

        # Set the number of rows and columns for the subplot
        num_cols = 2
        num_rows = round(len(metrics) / num_cols)
        nbins = 150

        # Create a figure and subplots
        fig, axes = plt.subplots(num_rows, num_cols, figsize=(20, 5 * num_rows))

        # Flatten the axes array
        axes = axes.flatten()

        # Iterate over the metrics and plot the histograms
        for i, metric in enumerate(metrics):
            # The first plot is clearer than the second - swap them as preferred
            axes[i].hist(graph_stats_1[metric], bins=nbins, alpha=0.5, label=label_1, log=True)

            axes[i].hist(graph_stats_2[metric], bins=nbins, alpha=0.5, label=label_2,  log=True)

            # Set tick parameters
            axes[i].tick_params(axis='both', which='major', labelsize=20)
            
            if metric in ['degree_t', 'betweenness', 'IVI']:
                axes[i].xaxis.set_major_locator(plt.MaxNLocator(15))

            # Set the title and labels for the subplot
            axes[i].set_xlabel(metric, fontsize=21)
            # axes[i].set_ylabel('count',fontsize=21)

            # Add a legend to the subplot
            axes[i].legend(prop={'size': 16})
    

        # Adjust the spacing between subplots
        plt.tight_layout()
        plt.subplots_adjust(hspace=0.3)

        if path:
            # Save the figure
            path = f"{path}/net_metrics_{label_1}_{label_2}.png"
            fig.savefig(path, dpi=300)


        # Show the plot
        return fig, graph_stats_1, graph_stats_2
    

    @staticmethod
    def plot_corr_matrix_coms(coms, height=700, title="Corr matrix", hide_up=False):

        # corr, pvalue = spearmanr(tum.exps["standard_4K_10TF"].mevsMut)
        coms.index.names = ["Mut_th"]
        dmy = coms.rename(columns={"All_genes": 0}).T
        dmy.columns = ["Com_{}".format(int(val)) for val in sorted(list(map(int, dmy.columns)))]

        corr, pvalue = spearmanr(dmy)

        # replace all the upper triangle values with None. So that we don't display the entire heatmap
        if hide_up:
            N = corr.shape[0]
            corr = [[corr[i][j] if i > j - 1 else None for j in range(N)] for i in range(N)]
            pvalue = [[pvalue[i][j] if i > j - 1 else None for j in range(N)] for i in range(N)]

        cols = dmy.columns
        corr_df = pd.DataFrame(corr, index=cols, columns=cols)

        # corr_df = dmy.corr(method="spearman")
        fig = px.imshow(corr_df, text_auto=True, height=height, title=title, aspect="auto")  # contrast_rescaling="infer" )

        return fig



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
@File    :   GraphToolExp.py
@Time    :   2023/07/11 09:06:37
@Author  :   Vlad Ungureanu
@Version :   1.0
@Contact :   vlad.ungureanu@york.ac.uk
@Desc    :   This file contains the class needed to analyse the graph tool
"""


import os
import pickle as pickle

# import _pickle as cPickle
import graph_tool.all as gt
import numpy as np
import pandas as pd
# ploting libs
import plotly.express as px
import plotly.graph_objects as go
from matplotlib import pyplot as plt
from plotly.subplots import make_subplots

# custom library
from .NetworkOutput import NetworkOutput
from .ExperimentSet import ExperimentSet



class GraphToolExperiment(NetworkOutput):

    com_df: pd.DataFrame
    gt_modCon: dict
    pos: any
    mevsMut: pd.DataFrame
    gt_g: gt.Graph
    sbm_method: str

    # TODO: Rename these to reflect that are a dictionary that contain information about BlockModelwh
    hstateObj: dict
    states: list

    def __init__(self, exp: NetworkOutput, rel_path=""):
        super().__init__(exp.graph, exp.exp_meta, exp.mut_df, exp.exps_path, rel_path)

    @classmethod
    def from_pgcna_exp(cls, exp: NetworkOutput, rel_path="../"):
        obj = cls.__new__(cls)
        super(GraphToolExperiment, obj).__init__(exp.graph, exp.exp_meta, exp.mut_df, exp.exps_path, rel_path=rel_path)
        obj._value = exp
        return obj

    @classmethod
    def from_pgcna_inet(cls, exp: NetworkOutput, rel_path="../"):
        obj = cls.__new__(cls)
        super(GraphToolExperiment, obj).__init__(exp.graph, exp.exp_meta, exp.mut_df, exp.exps_path, rel_path=rel_path, exp_type="iNet", base_path=exp.base_path)

        sbm_method = exp.type.split("_")[-1]
        gt_g_path = f"{exp.exps_path}/Networks/{exp.name}/EPG3/gt_{sbm_method}_{exp.name}.pickle"

        if os.path.isfile(gt_g_path):
            with open(gt_g_path, "rb") as handle:
                part_state = pickle.load(handle)
                if sbm_method == 'sbm': 
                    exp.states = [part_state]
                    gt_g = part_state['state'].g
                else:
                    exp.hstateObj = part_state
                    gt_g = part_state['state'].g
        else:
            raise ValueError("Couldn't loat gt tool {}".format(gt_g_path))

        exp.gt_g = gt_g
        obj._value = exp
        return obj

    # get the values from the original PGCNA experiment
    def __getattr__(self, name):
        if "_value" in self.__dict__:
            return getattr(self._value, name)
        raise AttributeError("no attribute: " + name)

    def create_gt(self):
        """
        Creates the Graph Tool Graph based on the list of edges generated with PGCNA
        """
        list_of_edges = list(self.edges_df[["Source", "Target", "Weight"]].itertuples(index=False, name=None))
        gt_g = gt.Graph(directed=False)
        vertex_prop = gt_g.add_edge_list(edge_list=list_of_edges, hashed=True, eprops=[("weight", "double")])
        gt_g.vp.gene = vertex_prop

        self.gt_g = gt_g

    def gt_modCon_MEV(self, all_tpms: pd.DataFrame, num_genes=100, is_imev=False, com_df = None, **kwargs):

        if not hasattr(self, "gt_modCon"):
            if com_df is not None:
                self.get_ModCon(com_df=com_df)
            else:
                if self.sbm_method == "sbm":
                    self.add_vp()
                else:
                    self.hsbm_add_vp()

                self.get_ModCon()

        is_mut = ("mut_df" in kwargs.keys())
        if (not is_imev) and (not is_mut):
            sort_col = "ModCon_{}_gt".format(self.type)
            self.mevsMut, _ = self.get_mevs(tpms=all_tpms, modCon=self.gt_modCon, sort_col=sort_col, num_genes=num_genes, verbose=False)
        elif (is_imev) and (not is_mut):
            sort_col = "ModCon_{}_gt".format(self.type)
            self.mevsMut, _ = self.get_iMevs(
                h_tpms=self.tpm_df, tum_tpms=all_tpms, modCon=self.gt_modCon, sort_col=sort_col, num_genes=num_genes, verbose=False)
        else:
            sort_col = "ModCon_{}_gt".format(self.type)
            self.mevsMut, _ = self.get_iMevs(h_tpms = self.tpm_df, tum_tpms=all_tpms, modCon=self.gt_modCon, sort_col=sort_col, num_genes=num_genes, verbose=False, mut_df=kwargs.pop('mut_df'), offset=kwargs.pop('offset'))

    # Sampling from posterior distrib
    def sample_posterior(self, n_iter=10000, mc_iter=10, deg_cor=True, distrib="real-exponential", verbose=True):
        gt_g = self.gt_g
        state = gt.BlockState(gt_g, recs=[gt_g.ep.weight], rec_types=[distrib], deg_corr=deg_cor)

        print("### Equilibrate MCMC for {}".format(self.type))
        gt.mcmc_equilibrate(state, wait=n_iter / 10, mcmc_args=dict(niter=mc_iter), verbose=verbose)

        bs = []  # collect some partitions
        h = np.zeros(gt_g.num_vertices() + 1)  # to find the probabilities for a certain size
        dls = []  # description length history

        def collect_partitions(s):
            bs.append(s.b.a.copy())
            B = s.get_nonempty_B()
            h[B] += 1
            dls.append(s.entropy())

        print("### Collecton partitions for {}".format(self.type))

        # Now we collect partitions for exactly 1000 sweeps, at intervals
        # of 10 sweeps:
        # n_iter = 10000
        eq_res = gt.mcmc_equilibrate(state, force_niter=n_iter, mcmc_args=dict(niter=mc_iter), callback=collect_partitions, verbose=verbose)

        print("### Marginal property partitions for {}".format(self.type))
        # Disambiguate partitions and obtain marginals
        pmode = gt.PartitionModeState(bs, converge=True, relabel=True)
        print("### get all marginal for {}".format(self.type))
        pv = pmode.get_marginal(gt_g)

        # keep track
        if not hasattr(self, "states"):
            self.states = []

        new_state = {
            "name": "state_{}".format(len(self.states) + 1),
            "partition_meta": {"h": h, "n_iter": n_iter, "bs": bs},
            "pmode": pmode,
            "state": state,
            "pv": pv,
            "dls": dls,
            "eq_res": eq_res,
        }

        self.states.append(new_state)

    def add_vp(self, state_idx=0, mut_df=None):
        gt_g = self.gt_g

        # check if pv and pmode exists
        if not hasattr(self, "states"):
            self.sample_posterior()

        state = self.states[state_idx]

        vp_marg = gt_g.new_vertex_property("object", state["pv"])
        gt_g.vp["marginal"] = vp_marg

        vp_max = gt_g.new_vertex_property("string", state["pmode"].get_max(gt_g))
        gt_g.vp["max_b"] = vp_max

        if mut_df is not None:
            comb = {}
            for idx in gt_g.vertex_index:
                comb[idx] = gt_g.vp.gene[idx]

            dmy_df = pd.DataFrame.from_dict(comb, orient="index", columns=["gene"]).reset_index(names="gt_idx").set_index("gene")
            dmy_df["mut_count"] = mut_df["count"]
            dmy_df.fillna(0, inplace=True)
            dmy_df["mut_count"] = dmy_df["mut_count"].astype(int)
            dmy_df.sort_values(by="gt_idx", inplace=True)

            vp_mut = gt_g.new_vertex_property("int", list(dmy_df["mut_count"].values))
            gt_g.vp["mut_count"] = vp_mut

    # Layout functions
    def layout(self):
        pos = gt.sfdp_layout(self.gt_g, eweight=self.gt_g.ep.weight)
        vb, eb = gt.betweenness(self.gt_g)
        self.pos = pos

    # overriding the method from the PGCNAOutput
    def get_ModCon(self, state=0, com_df = None):
        if self.sbm_method == 'sbm':
            if com_df is None:
                com_df = self.get_gt_df(state_idx=state)
        else:
            if com_df is None:
                com_df, _ = self.hsbm_get_gt_df()
            com_df["max_b"] = com_df["P_lvl_0"]

        gen_coms = com_df["max_b"].reset_index().rename(columns={"index": "Id"})
        mut_df = self.mut_df
        meta_df = self.meta_df

        # create a DataFrame with the components for the modCon equation
        modifier = self.type.split("_")[0]
        col = "conn_{}".format(modifier)

        # all_genes = []  # can be used for testing
        modCons = {}
        for mod_class in gen_coms["max_b"].unique():
            genes = gen_coms.loc[gen_coms["max_b"] == mod_class]["Id"].values

            conn_g = []
            for gene in genes:
                # 1st condition, get all the edges that the current gene is the Source and are inside the module
                # 2nd condition, get all the edges that current is the Target and are inside the module
                in_edges = self.edges_df[(self.edges_df["Source"] == gene) | (self.edges_df["Target"] == gene)]
                # Removed as a way to deal with communities where nodes have degree_in = 0
                # in_edges = in_edges[in_edges["Source"].isin(genes)]
                # in_edges = in_edges[in_edges["Target"].isin(genes)]
                weights_sum = in_edges["Weight"].sum()
                if weights_sum == 0:
                    print(f"Weighted sum = 0 for Com {mod_class} !!")
                conn_g.append([gene, weights_sum])

            conn_df = pd.DataFrame(conn_g, columns=["gene", col]).set_index("gene")
            working_df = pd.concat([conn_df, meta_df.loc[meta_df["genes"].isin(genes)].set_index("genes"), mut_df[mut_df.index.isin(genes)]["count"]], axis=1)

            # 5. Workout the ModCon and save it
            working_df["ModCon_{}_gt".format(self.type)] = (
                (working_df[col] ** 2) * working_df["q2E"] * working_df["varWithin"] * (100 - working_df["varAcross"]) / 100
            )

            # sort by the current exp type
            modCons[mod_class] = working_df.sort_values(by="ModCon_{}_gt".format(self.type), ascending=False)

        self.gt_modCon = modCons
        return modCons

    # Experiment processing
    def get_gt_df(self, state_idx=0, compute=True):
        """
        The graph requires the following VertexPropertyMaps:
        * Marginal - marginal prob of each node
        * Max_b - the max partition prob of each node

        Args:
            gt_g (Graph): Graph-Tool Graph
            pmode (PartitionModeState): _description_
        """

        # check if there is already calculated
        if not compute and hasattr(self, "com_df"):
            return self.com_df

        gt_g = self.gt_g

        # get the maximum number of clusters
        state = self.states[state_idx]
        max_part = state["pmode"].get_B()
        num_nodes = gt_g.num_vertices()

        part_nodes = np.zeros((num_nodes, max_part + 2))
        genes = []
        for node in gt_g.iter_vertices():
            genes.append(gt_g.vp.gene[node])
            frac = np.array(gt_g.vp["marginal"][node])
            part_nodes[node, : frac.shape[0]] = frac
            part_nodes[node, -2] = node
            part_nodes[node, -1] = int(gt_g.vp["max_b"][node]) # offset to start from 0

        col_names = ["b_{}".format(idx) for idx in range(0, max_part)] + ['node_idx', "max_b"]
        com_df = pd.DataFrame(part_nodes, columns=col_names, index=genes)

        self.com_df = com_df
        return com_df

    def check_graph(self):
        """

        TODO Work in progress. Re-construct the edges_df matrix but from gt_g.

        It just needs to check if the two are identical
        """
        if not self.gt_g:
            self.create_gt()

        gt_g = self.gt_g
        for e in enumerate(gt_g.iter_edges()):
            source, target = gt_g.vp.gene[e[0]], gt_g.vp.gene[e[1]]
            weight = gt_g.ep.weight[(e[0], e[1])]
            print("{} - {}: {}".format(source, target, weight))

    def get_stable_genes(self, state_idx=0, prob_th=0.75):
        self.add_vp(state_idx=state_idx)
        com_df = self.get_gt_df(state_idx=state_idx)

        max_iters = com_df.max().max()
        th = max_iters * prob_th

        stable_genes = com_df.loc[com_df[com_df > th].dropna(how="all").index]
        com_df["stable_gene"] = "wobble"
        com_df.loc[stable_genes.index, "stable_gene"] = "stable"

        return com_df

    ### hsbm specific methods
    def hsbm_get_gt_df(self):
        """
        The graph requires the following VertexPropertyMaps:
        * Marginal - marginal prob of each node
        * Max_b - the max partition prob of each node

        Args:
            gt_g (Graph): Graph-Tool Graph
            pmode (PartitionModeState): _description_
        """
        gt_g = self.gt_g

        # get the maximum number of clusters
        max_part = self.hstateObj["pmode"].get_B()
        num_nodes = gt_g.num_vertices()

        # 2 for max-b and node_idx
        part_nodes = np.zeros((num_nodes, max_part + 2))
        genes = []
        for node in gt_g.iter_vertices():
            genes.append(gt_g.vp.gene[node])
            frac = np.array(gt_g.vp["marginal"][node])
            part_nodes[node, : frac.shape[0]] = frac
            # Last two position with the node idx and the partition with highest membership
            part_nodes[node, -2] = node
            part_nodes[node, -1] = int(gt_g.vp["max_b"][node]) # offset to start from 0

        col_names = ["b_{}".format(idx) for idx in range(0, max_part)] + ["node_idx", "level_0"]
        com_df = pd.DataFrame(part_nodes, columns=col_names, index=genes)

        #### adding the gene membership from all the levels

        # Get the non-1 levels
        non_zero_lvls = []
        levels = self.hstateObj["state"].get_levels()
        for idx, s in enumerate(levels):
            if s.get_N() == 1:
                break
            non_zero_lvls.append(idx)

        # create DF with the non-1 membersips
        lvl_parts = self.get_level_partitions(com_df)

        # check if the dataframes are alligned before mergin
        if not lvl_parts["P_lvl_0"].equals(com_df["level_0"].astype(int)):
            raise ValueError("The partition membership don't match, between the two dataframes")

        # merge
        com_df = pd.concat([com_df.drop(columns=["level_0"]), lvl_parts], axis=1)
        com_df["max_b"] = com_df["P_lvl_0"]

        return com_df, lvl_parts.columns

    def get_level_partitions(self, com_df):
        levels = self.hstateObj["state"].get_levels()
        # create the dictionary
        lvls_parts = {"gene": []}
        for idx, lvl in enumerate(levels):
            lvls_parts[idx] = []
            if lvl.get_N() == 1:
                break

        # add values
        for gene, row in com_df.iterrows():
            # first, will be from the first level
            lvls_parts["gene"].append(gene)
            last_lvl_idx = row["node_idx"]
            for i, lvl in enumerate(levels):
                part = lvl.get_blocks()[last_lvl_idx]
                lvls_parts[i].append(part)
                last_lvl_idx = part
                if lvl.get_N() == 1:
                    break

        dmy = pd.DataFrame.from_dict(data=lvls_parts)
        dmy.columns = ["gene"] + list(["P_lvl_{}".format(lvl) for lvl in range(0, len(lvls_parts.keys()) - 1)])

        return dmy.set_index("gene")

    def hsbm_sample_posterior(self, n_iter=10000, mc_iter=10, deg_cor=False):
        gt_g = self.gt_g
        # TODO: Is this the best function?
        state = gt.NestedBlockState(gt_g, recs=[gt_g.ep.weight], rec_types=["real-exponential"], state_args=dict(deg_corr=deg_cor))

        print("### 1. Equilibrate MCMC. n_iter {}".format(n_iter / 10))
        # We will first equilibrate the Markov chain
        gt.mcmc_equilibrate(state, wait=n_iter / 10, mcmc_args=dict(niter=mc_iter), verbose=True)  # 1000 / 10 standard

        bs = []  # collect some partitions
        h = [np.zeros(gt_g.num_vertices() + 1) for s in state.get_levels()]
        dls = [np.zeros(gt_g.num_vertices() + 1) for s in state.get_levels()]

        # to find the probabilities for a certain size
        def collect_partitions(s):
            for l, sl in enumerate(s.get_levels()):
                B = sl.get_nonempty_B()
                h[l][B] += 1
            dls.append(s.entropy())
            bs.append(s.get_bs())

        print("\n\n\n### 2. Collecton partitions. n_iter {}".format(n_iter))
        eq_res = gt.mcmc_equilibrate(state, force_niter=n_iter, mcmc_args=dict(niter=mc_iter), callback=collect_partitions, verbose=True)  # 10000 / 10

        print("### 3. Marginal property partitions")
        # Disambiguate partitions and obtain marginals
        pmode = gt.PartitionModeState(bs, nested=True, converge=False)
        print("### 4. Collecting the data")
        pv = pmode.get_marginal(gt_g)

        # Get consensus estimate
        bs = pmode.get_max_nested()
        state = state.copy(bs=bs)

        print("### 5. Saving")
        self.hstateObj = {
            "name": "hstate",
            "partition_meta": {"h": h, "n_iter": n_iter, "bs": bs},
            "pmode": pmode,
            "state": state,
            "pv": pv,
            "dls": dls,
            "eq_res": eq_res,
        }

    def draw_hstate(self, name, level=None, lvl_idx=0, layout="radial"):
        if not hasattr(self, "hstateObj"):
            raise ValueError("No hstate")

        if level is None:
            filename = "hbm_{}_{}_{}_marginals_2.pdf".format(self.type, layout, name)
            self.hstateObj["state"].draw(
                layout=layout,
                vertex_text=self.gt_g.vp.gene,
                vertex_shape="pie",
                vertex_pie_fractions=self.hstateObj["pv"],
                # edge_color=gt.prop_to_size(self.gt_g.ep.weight, power=1, log=True),
                # eorder=self.gt_g.ep.weight, edge_pen_width=gt.prop_to_size(self.gt_g.ep.weight, 1, 4, power=1, log=True),
                # order=self.gt_g.ep.weight,
                # edge_pen_width=gt.prop_to_size(self.gt_g.ep.weight, 2, 8, power=1, log=True),
                # edge_gradient=[],
                output=filename,
            )
        else:
            filename = "hbm_lvl{}_{}_{}_{}_marginals.pdf".format(lvl_idx, self.type, layout, name)
            level.draw(output=filename)

    def print_summary(self):
        if not hasattr(self, "hstateObj"):
            raise ValueError("No hstate")

        self.hstateObj["state"].print_summary()

    def hsbm_add_vp(self, mut_df=None):
        # check if pv and pmode exists
        if not hasattr(self, "hstateObj"):
            raise ValueError("No hstate")

        gt_g = self.gt_g

        comb = []
        for idx in gt_g.vertex_index:
            comb.append({idx: gt_g.vp.gene[idx]})

        stateObj = self.hstateObj
        vp_meta = gt_g.new_vertex_property("object", comb)
        gt_g.vp["meta"] = vp_meta

        vp_marg = gt_g.new_vertex_property("object", stateObj["pv"])
        gt_g.vp["marginal"] = vp_marg

        vp_max = gt_g.new_vertex_property("int", stateObj["pmode"].get_max(gt_g))
        gt_g.vp["max_b"] = vp_max

        if mut_df is not None:
            comb = {}
            for idx in gt_g.vertex_index:
                comb[idx] = gt_g.vp.gene[idx]

            dmy_df = pd.DataFrame.from_dict(comb, orient="index", columns=["gene"]).reset_index(names="gt_idx").set_index("gene")
            dmy_df["mut_count"] = mut_df["count"]
            dmy_df.fillna(0, inplace=True)
            dmy_df["mut_count"] = dmy_df["mut_count"].astype(int)
            dmy_df.sort_values(by="gt_idx", inplace=True)

            vp_mut = gt_g.new_vertex_property("int", list(dmy_df["mut_count"].values))
            gt_g.vp["mut_count"] = vp_mut

    def hsbm_plot_posterior(self):
        """
        Plots the posterior probability for each layer of hSBM until only a single block is found by the highest probability

        Args:
            stop_lvl (int, optional): The level at which to stop plotting. Defaults to -1 so it doesn't stop at all

        Returns:
            None
        """
        if not hasattr(self, "hstateObj"):
            raise ValueError("No hstate")

        h = self.hstateObj["partition_meta"]["h"]
        n_iter = self.hstateObj["partition_meta"]["n_iter"]
        levels: gt.NestedBlockState = self.hstateObj["state"].get_levels()

        figs, titles = [], []
        for l, lvl_state in enumerate(levels):
            cluster_idxs = np.nonzero(h[l])[0]

            cluster_contrib = h[l][cluster_idxs] / n_iter
            cluster_idxs = list(map(str, cluster_idxs))
            figs.append(px.bar(x=cluster_idxs, y=cluster_contrib))
            titles.append("Layer {}".format(l))
            print(f"Lvl {l}, non-empty {lvl_state.get_nonempty_B()}; B {lvl_state.get_B()}; N: {lvl_state.get_N()}")

            if lvl_state.get_nonempty_B() == 1:
                break

        if figs == []:
            print("No figures to plot")
            return 

        config = {
            "num_cols": 4,
            "shared_x": False,
            "shared_y": True,
            "h_spacing": 0.05,
            "v_spacing": 0.15,
            "main_title": "Posterior probability for each layer of hSBM",
            "height": 900,
            "width": None,
            "y_title": "Y-axis",
            "x_title": "X-axis",
            "specs": None,
        }

        return GraphToolExperiment.helper_multiplots(figs, titles, config)

    # TODO: Develop this to accept gene name
    def get_gene_membership_levels(self, gene):
        levels = self.hier_state.get_levels()
        for s in levels:
            print(s)
            if s.get_N() == 1:
                break

        r = levels[0].get_blocks()[46]  # group membership of node 46 in level 0
        print(r)
        r = levels[1].get_blocks()[r]  # group membership of node 46 in level 1
        print(r)
        r = levels[2].get_blocks()[r]  # group membership of node 46 in level 2
        print(r)

    # TODO: Work in progress
    def model_selection(self):
        g = gt.collection.ns["foodweb_baywet"]

        # This network contains an internal edge property map with name
        # "weight" that contains the energy flow between species. The values
        # are continuous in the range [0, infinity].

        state = gt.minimize_nested_blockmodel_dl(g, state_args=dict(recs=[g.ep.weight], rec_types=["real-exponential"]))

        # improve solution with merge-split

        for i in range(100):
            ret = state.multiflip_mcmc_sweep(niter=10, beta=np.inf)

        state.draw(
            edge_color=gt.prop_to_size(g.ep.weight, power=1, log=True),
            ecmap=(plt.cm.inferno, 0.6),
            eorder=g.ep.weight,
            edge_pen_width=gt.prop_to_size(g.ep.weight, 1, 4, power=1, log=True),
            edge_gradient=[],
            output="foodweb-wsbm.pdf",
        )

    # Plotting functionsget
    def plot_partition_prob(self, state_idx=0):
        if not hasattr(self, "states"):
            raise AttributeError("No attribute partitions_meta")

        part_meta = self.states[state_idx]["partition_meta"]
        cluster_idxs = np.nonzero(part_meta["h"])[0]
        cluster_contrib = part_meta["h"][cluster_idxs] / part_meta["n_iter"]
        cluster_idxs = list(map(str, cluster_idxs))
        fig = px.bar(x=cluster_idxs, y=cluster_contrib, title="Marginal prob of the number of groups")

        return fig

    def plot_membership_changes(self, idx, prob_th=90):
        com_df = self.get_gt_df(state_idx=idx)

        # wobble df defined by <75
        max_iters = com_df.max().max()
        th = max_iters * prob_th

        wobble_df = com_df[com_df < th].dropna().copy(deep=True)
        dmy_df = pd.concat(
            [pd.DataFrame(wobble_df["max_b"].value_counts()), pd.DataFrame(com_df["max_b"].value_counts()).rename(columns={"count": "orig_size"})], axis=1
        ).sort_index()

        fig = px.line(dmy_df.reset_index(), x="max_b", y=["count", "orig_size"], markers=True)
        fig.update_layout(xaxis=dict(tickmode="linear", tick0=0, dtick=1))

        return fig

    def comp_posterior_distrib(self):
        figs, titles = [], []
        for idx, _ in enumerate(self.states):
            figs.append(self.plot_partition_prob(state_idx=idx))
            titles.append("Run {}".format(idx + 1))

        subplots_config = {
            "num_cols": 2,
            "shared_x": False,
            "shared_y": True,
            "h_spacing": 0.05,
            "v_spacing": 0.15,
            "main_title": "SBM #communities for exp {} in {} runs.".format(self.type, len(titles)),
            "height": 700,
            "width": None,
            "y_title": "#Comunities",
            "x_title": "Probability",
            "specs": None,
        }

        return GraphToolExperiment.helper_multiplots(figs, titles, subplots_config)

    # NOTE this only works for SBM
    def plot_genes_network_metrics(self, state_idx=0, prob_th=0.75):
        # get the stable genes based on the prob_th
        com_df = self.get_stable_genes(state_idx=state_idx, prob_th=prob_th)

        # get the iGraph stats
        graph_stats = self.compute_graph_stats()

        # The other metrics
        metrics = ["TF", "count", "IVI", "degree", "hubScore", "strength", "betwenees"]
        figs = []
        for metric in metrics:
            if metric in graph_stats.columns:
                com_df[metric] = graph_stats[metric]
            else:
                if metric != "TF":
                    # Need to make TF as category, to show nicely on the histogram
                    com_df[metric] = self.nodes_df[metric]
                else:
                    com_df[metric] = self.nodes_df[metric].astype(str)

            figs.append(px.histogram(com_df, x=metric, color="stable_gene", nbins=None, histnorm="percent"))

        # prepare for plotting
        titles = metrics

        subplots_config = {
            "num_cols": 3,
            "shared_x": False,
            "shared_y": False,
            "h_spacing": 0.03,
            "v_spacing": 0.1,
            "main_title": "Stable genes against network metrics. Network - {}. State_idx - {}".format(self.type, state_idx),
            "height": 900,
            "width": None,
            "y_title": None,
            "x_title": None,
            "specs": None,
        }

        return figs, titles, subplots_config

    # Network stats
    def compute_graph_stats(self):
        graph_stats = pd.DataFrame(index=[g for g in self.graph.vp["gene"]])

        # Degree related metrics
        vp = self.graph.degree_property_map(deg="total", weight=self.graph.ep["weight"])
        graph_stats["degree_w"] = np.round(vp.a, 5)

        vp = self.graph.degree_property_map(deg="total")
        graph_stats["degree_t"] = vp.a

        vp_bet, _ = gt.betweenness(self.graph, weight=self.graph.ep["weight"])
        graph_stats["betweenness"] = [g for g in vp]

        # Other centrality stats
        vp = gt.closeness(self.graph, weight=self.graph.ep["weight"])
        graph_stats["closeness"] = np.round(vp.a, 5)

        vp = gt.katz(self.graph, weight=self.graph.ep["weight"])
        graph_stats["katz"] = np.round(vp.a, 5)

        _, vp, _ = gt.hits(self.graph, weight=self.graph.ep["weight"])
        graph_stats["hits"] = vp.a

        vp = gt.central_point_dominance(self.graph, betweenness=vp_bet)
        graph_stats['central_point_dominance'] = round(vp, 5)

        # More degree metrics
        vp = self.graph.degree_property_map(deg="in", weight=self.graph.ep["weight"])
        graph_stats["degree_i"] = vp.a

        vp = self.graph.degree_property_map(deg="out", weight=self.graph.ep["weight"])
        graph_stats["degree_o"] = vp.a

        vp = gt.pagerank(self.graph)
        graph_stats["pageRank"] = vp.a

        return graph_stats

    # Overriding the one from PGCNA ouput
    def ge_comm(self):
        comm_tpms = {}
        median_vals, std_df = pd.DataFrame(index=self.tpm_df.columns), pd.DataFrame(index=self.tpm_df.columns)

        for mod_class in self.nodes_df['max_b'].unique():
            genes = self.nodes_df.loc[self.nodes_df['max_b'] == mod_class].index.values
            comm_tpms[mod_class] = self.tpm_df.loc[self.tpm_df.index.isin(genes)]

            median_vals = pd.concat([median_vals, comm_tpms[mod_class].median().rename("Comm_{}".format(mod_class))], axis=1)
            std_df = pd.concat([std_df, comm_tpms[mod_class].std().rename("Comm_{}".format(mod_class))], axis=1)

        med_df = median_vals.melt(var_name="Comm", value_name="Median", ignore_index=False).reset_index(names=["Sample"])
        std_df = std_df.melt(var_name="Comm", value_name="Std", ignore_index=False).reset_index(names=["Sample"])
        combined_stats = pd.concat([std_df, med_df["Median"]], axis=1)
        return combined_stats

    ######## Adding custom properties ########
    def add_gt_prop_draw(self, gt_g: gt.Graph, com_df: pd.DataFrame, tf_list: list):

        if not hasattr(self, "nodes_df"):
            self.export_to_gephi(save=False, com_df=com_df)

        self.nodes_df["ModCon_Rank"] = self.nodes_df["ModCon_Rank"].astype(int)
        rank_th = 50
        hex_colors = GraphToolExperiment.color_scale(color_scale="blues", num_points=rank_th)

        max_char, prcsd_genes, display_prop, is_tf, colors_rank = (6, [], [], [], [])
        for idx in gt_g.vertices():
            gene, mut_count, com = gt_g.vp["gene"][idx], gt_g.vp["mut_count"][idx], gt_g.vp['max_b'][idx]
            rank = self.nodes_df.loc[gene]["ModCon_Rank"]
            if rank > rank_th:
                colors_rank.append("#a2a1a1")
                # colors_rank.append("#5A54A0") - purple the highest color in Sunset scale
            else:
                colors_rank.append("#c86514")
                # colors_rank.append(hex_colors[rank-1])

            if len(gene) > max_char:
                if "ENSG" in gene:
                    gene = "EN_" + gene[-3:]
                else:
                    gene = gene[:max_char] + "."

            if gene in tf_list:
                is_tf.append(1)
            else:
                is_tf.append(0)

            prcsd_genes.append(gene)
            display_prop.append(f'{gt_g.vp["gene"][idx]}: Mut {mut_count}, Com: {com}')

        # vp_modCon_rank = gt.new_vertex_proeprty('int', modCon_rank)
        # gt_g.vp['modCon_rank'] = vp_modCon_rank
        # vertex_fill_color=gt.prop_to_size(vb, 0, 1, power=.1)  

        vp_prcsd_gene = gt_g.new_vertex_property("string", prcsd_genes)
        gt_g.vp["prcsd_gene"] = vp_prcsd_gene

        vp_display_prop = gt_g.new_vertex_property("string", display_prop)
        gt_g.vp["display_prop"] = vp_display_prop

        vp_colors_rank = gt_g.new_vertex_property("string", colors_rank)
        gt_g.vp["colors_rank"] = vp_colors_rank

        vp_tf = gt_g.new_vertex_property("int", is_tf)
        gt_g.vp["is_tf"] = vp_tf

    ######## Helper methods ########
    def save(self, exp_type="sbm", is_hsbm=False, path=None):
        save_gt_exp = {"name": self.name, "gt_exp": self}
        if path is None:
            path = self.exps_path

        # folder location
        if is_hsbm:
            folder_path = f"{path}/hsbm/{exp_type}/"
        else:
            folder_path = f"{path}/sbm/{exp_type}/"

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

        filename = f"gt_{exp_type}_{self.name}.pickle"
        full_path = folder_path + filename
        with open(full_path, "wb") as handle:
            pickle.dump(save_gt_exp, handle, protocol=pickle.HIGHEST_PROTOCOL)
        return full_path

    def export_gephi_coord(self, graph: gt.Graph, draw_results: tuple, path: str, name: str):

        test_array = np.array(list(draw_results[0]))

        x_cord = test_array[:, 0] * 40
        y_cord = test_array[:, 1] * 40

        vp_x = graph.new_vertex_property("double", x_cord)
        graph.vp["x_cord"] = vp_x
        graph.vp["x"] = vp_x

        vp_y = graph.new_vertex_property("double", y_cord)
        graph.vp["y_cord"] = vp_y
        graph.vp["y"] = vp_y

        graph.save(f"{path}/{name}.graphml")

    def export_morpheus(self, settings: dict):

        cs_num_1 = settings['cs_num_1']
        cs_num_2 = settings['cs_num_2']
        sel_mut = settings['sel_mut']
        cs_exp = settings['cs_exp']
        vu_output = settings['vu_output']
        label_col_cs = settings['label_col_cs']

        file_path= f"{settings['figures_path']}/{settings['filename']}.tsv"

        # MEVS
        mevsMut = self.mevsMut

        cs_cols = [f"RawKMeans_CS_{cs_num_1}_{label_col_cs}", f"RawKMeans_CS_{cs_num_2}_{label_col_cs}"]
        vu_cols = [
            "race",
            "gender",
            "bmi",
            "KMeans_labels_6",
            "2019_consensus_classifier",
            "TCGA408_classifier",
            "Lund2017.subtype"
        ]

        # Selecting only the present mutations
        dmy_mut = sel_mut[sel_mut.index.isin(self.tpm_df.index)]

        mevs_cols, mut_cols = list(self.mevsMut.columns), list(dmy_mut.T.columns)
        mevs_cs = pd.concat(
            [
                cs_exp[cs_cols],
                vu_output[vu_cols],
                dmy_mut.T,
                mevsMut,
            ],
            axis=1,
        ).dropna()

        if 'diff_type' in settings.keys():
            mevs_cs = pd.concat(
                [
                    settings['diff_type']["Diff Type"],
                    mevs_cs.T,
                ],
                axis=1,
            ).T

        for col in cs_cols:
            mevs_cs[col] = mevs_cs[col].astype(str)

        mevs_cs = mevs_cs[vu_cols + cs_cols + mut_cols + mevs_cols]

        # Rename for a better aspect
        remap_cols = {"TCGA408_classifier": "TCGA", "KMeans_labels_6": "CA + IFNG", "2019_consensus_classifier": "Consensus",  "Lund2017.subtype": "Lund"}
        mevs_cs = mevs_cs.rename(columns=remap_cols)

        mevs_cs.transpose().to_csv(f"{file_path}", sep="\t", index=True)

        return mevs_cs

    ### Looking at the nodes ###
    def find_node(self, gene: str):
        for v in self.gt_g.iter_vertices():
            if self.gt_g.vp["gene"][v] == gene:
                return v, self.gt_g.vp['max_b'][v]

    def filter_graph(self, gene: str, show_own_com=True, verbose=True):

        graph = self.gt_g
        # Clear the previous filter
        graph.set_edge_filter(None)
        graph.set_vertex_filter(None)

        # setup the flag 
        vp_bool = graph.new_vertex_property("bool")
        neighbors = []        

        # idx, non_zero_part = find_comm_membership(gt_g, gene=gene, n_runs=10000)
        gene_idx, com = self.find_node(gene=gene)

        if verbose:
            print(f"{gene}. Idx = {gene_idx}. Comm = {com}")

        for idx in graph.iter_out_neighbors(gene_idx):
            vp_bool[idx] = True
            neighbors.append(idx)

        # Show the community where is the gene we seek
        if show_own_com:
            max_b = np.array(list(graph.vp["max_b"])) #
            # For hSBM and newer exp it works graph.vp["max_b"] 
            #  but for SBM and the control experiments it doesn't 
            #  so I had to explicitly transfer the VP of max_b in a numpy array
            for idx in np.nditer(np.where(max_b == com)):
                vp_bool[idx] = True

        vp_bool[gene_idx] = True
        graph.set_vertex_filter(vp_bool)

        return neighbors

    def get_gene_neigbhors(self, gene_name="BNC1", verbose=True):

        neighbors = self.filter_graph(gene=gene_name, verbose=verbose)
        self.gt_g.set_vertex_filter(None)

        neighbors_stats, neighbors_df = self.process_neigbhors(neighbors)
        return neighbors_stats, neighbors_df
    
    # Show only one or more community
    def show_comms(self, communities: []):
        """
        Filters the graph to show only the vertices belonging to the specified communities.
        Only works for SBM and hSBM

        Parameters:
            communities (list): A list of community labels.

        Returns:
            None
        """
        graph = self.gt_g
        max_b = np.array(list(graph.vp["max_b"]))

        vp_bool = graph.new_vertex_property("bool")
        for com in communities:
            for idx in np.nditer(np.where(max_b == com)):
                vp_bool[idx] = True

        graph.set_vertex_filter(vp_bool)

    def process_neigbhors(self, v_idxs:list):

        nodes_df = self.nodes_df
        neigbhors_df = nodes_df[nodes_df["node_idx"].isin(v_idxs)][["max_b", "count", "TF"]]

        stats_neigbhor = []
        neighbors_genes = pd.DataFrame()

        for com in neigbhors_df["max_b"].unique():
            sel_com = neigbhors_df[neigbhors_df["max_b"] == com]
            com_size = nodes_df[nodes_df["max_b"] == com].shape[0]
            mut_burden = sel_com[sel_com["count"] > 0].shape[0]
            num_tf = sel_com[sel_com["TF"] == 1].shape[0]

            # com rep
            com_ratio = round(sel_com.shape[0] / com_size, 4)

            stats_neigbhor.append((com, com_ratio, mut_burden, num_tf, sel_com.shape[0], com_size))
            neighbors_genes = pd.concat([neighbors_genes, sel_com], axis=0)

        stats_neigbhor = pd.DataFrame(stats_neigbhor, columns=["com", "com_ratio", "mut_burden", "num_TF", "num_found", "com_size"])
        stats_neigbhor = stats_neigbhor.sort_values(by=["num_found"], ascending=False)
        stats_neigbhor["com"] = stats_neigbhor["com"].astype(str)

        return stats_neigbhor, neighbors_genes

    def plot_overview_gene(self, gene_name: str):

        neigbhors = self.filter_graph(gene=gene_name)
        self.gt_g.set_vertex_filter(None)
        stats_neigbhors, _ = self.process_neigbhors(v_idxs=neigbhors)

        fig1 = px.bar(stats_neigbhors, x="com", y=["num_found", "com_size"], barmode="group", text_auto=True)
        fig2 = px.bar(stats_neigbhors, x="com", y="com_ratio", barmode="group", text_auto=True)
        fig3 = px.bar(stats_neigbhors, x="com", y=["mut_burden"], barmode="group", text_auto=True)
        fig4 = px.bar(stats_neigbhors, x="com", y="num_TF", barmode="group", text_auto=True)

        figs = [fig1, fig2, fig3, fig4]
        titles = [
            "Num Found and Community sizes",
            "Com ratio representation",
            "Mut burden in the neigbhors",
            "Number of TF in the neigbhors",
        ]

        subplots_config = {
            "num_cols": 2,
            "shared_x": False,
            "shared_y": False,
            "h_spacing": 0.05,
            "v_spacing": 0.15,
            "main_title": f"Neigbhours overview for {gene_name} ({self.type})",
            "height": 700,
            "width": None,
            "y_title":None,
            "x_title": "Community",
            "specs": None,
        }

        fig = self.helper_multiplots(figs, titles, subplots_config)

        return fig


    @staticmethod
    def load_hsbm_exps(exp_set: ExperimentSet):

        exps = {}
        entropy = pd.DataFrame()
        for _, exp in enumerate(exp_set.get_exps()):
            print(f"Loading Graph-Tool for {exp.type}")

            tf = NetworkOutput.extract_tf_number(exp.name)
            exps[exp.type] = GraphToolExperiment.from_pgcna_inet(exp, rel_path="")
            exps[exp.type].export_to_gephi(save=False)

            # compute the entropy
            exps[exp.type].hsbm_add_vp()

            gt_state: gt.NestedBlockState = exp.hstateObj["state"]
            dls = exp.hstateObj['dls']
            lvls = len(gt_state.levels)

            dmy_df = pd.DataFrame(dls[lvls:], columns=["Entropy"])
            dmy_df["TF"] = tf
            entropy = pd.concat([entropy, dmy_df], axis=0)

        exp_set.exps = exps
        return exps, entropy
        

    # TODO: remove this, it is not generalisable
    @staticmethod
    def load_sbm_exps(exp_set, name="standard_4K", exp_type="h42", tf_range=None, is_hsbm=False):
        exps = {}
        entropy = pd.DataFrame()

        if tf_range is None:
            tf_range = range(3, 10, 1)

        for tf in tf_range:
            key = str(tf)
            sel_exp = exp_set.exps["{}_{}TF".format(name, tf)]
            exps[key] = GraphToolExperiment.load(exp_name=sel_exp.name, exp_type=exp_type, path=sel_exp.exps_path, is_hsbm=is_hsbm)

            exps[key].add_vp()
            dmy_df = pd.DataFrame(exps[key].states[0]["dls"], columns=["Entropy"])
            dmy_df["TF"] = key
            entropy = pd.concat([entropy, dmy_df], axis=0)

        return exps, entropy

    @staticmethod
    def load(exp_name, exp_type="sbm", is_hsbm=False, path="", legacy_path=True):
        filename = "gt_{}_{}.pickle".format(exp_type, exp_name)

        if legacy_path:
            if is_hsbm:
                full_path = "{}/hsbm/{}/{}".format(path, exp_type, filename)
            else:
                full_path = "{}/sbm/{}/{}".format(path, exp_type, filename)
        else:
            full_path = os.path.abspath(path)

        # full_path = os.path.abspath(full_path)
        if os.path.isfile(full_path):
            with open(full_path, "rb") as handle:
                exp = pickle.load(handle)
                # we need to update the experiment's path with the current one
                exp["gt_exp"].exps_path = path
                print("### Loaded {}".format(exp["name"]))
                return exp["gt_exp"]
        else:
            raise ValueError("SBM exp not found at {}".format(full_path))

    @staticmethod
    def helper_multiplots(figs, subplot_tiles, config=None):
        if config is None:
            config = {
                "num_cols": 2,
                "shared_x": False,
                "shared_y": True,
                "h_spacing": 0.05,
                "v_spacing": 0.15,
                "main_title": "Subplots",
                "height": 700,
                "width": None,
                "y_title": "Y-axis",
                "x_title": "X-axis",
                "specs": None,
            }

        num_cols = config["num_cols"]
        num_rows = int(np.ceil(len(figs) / num_cols))

        subplot = make_subplots(
            rows=num_rows,
            cols=num_cols,
            subplot_titles=subplot_tiles,
            shared_xaxes=config["shared_x"],
            shared_yaxes=config["shared_y"],
            horizontal_spacing=config["h_spacing"],
            vertical_spacing=config["v_spacing"],
            specs=config["specs"],
        )

        idx_row, idx_col = 1, 1
        for i, fig in enumerate(figs):
            for trace in range(len(fig["data"])):
                subplot.add_trace(fig["data"][trace], row=idx_row, col=idx_col)

            if idx_col % num_cols == 0:
                idx_col = 0
                idx_row += 1
            idx_col += 1

        layout = go.Layout(title_text=config["main_title"])

        subplot = subplot.update_layout(layout, height=config["height"], width=config["width"], showlegend=False)
        subplot = subplot.update_xaxes(title_text=config["x_title"])
        subplot = subplot.update_yaxes(title_text=config["y_title"])

        # Remove duplicate legend items
        # visited = []
        # for trace in subplot["data"]:
        #     if trace["name"] is None:
        #         continue

        #     trace["name"] = trace["name"].split(",")[0]
        #     if trace["name"] not in visited:
        #         trace["showlegend"] = True
        #         visited.append(trace["name"])
        #     else:
        #         trace["showlegend"] = False

        return subplot

    @staticmethod
    def create_mpl_fig(com_df: pd.DataFrame, size=6000, dpi=300):
        """
        Create a matplotlib figure with an empty plot and legend based on the provided DataFrame.

        Parameters:
        - com_df (pd.DataFrame): The DataFrame containing the data for plotting. This needs to contain 'max_b' column
        - size (int): The size of the figure in pixels (default: 6000).
        - dpi (int): The DPI (dots per inch) of the figure (default: 300).

        Returns:
        - fig (matplotlib.figure.Figure): The created matplotlib figure.
        - ax (matplotlib.axes.Axes): The axes of the created figure.
        """

        import matplotlib.cm

        # matplotlib.use("cairo")
        # Combine colors from two color maps
        default_clrs = list(matplotlib.cm.tab20.colors) + list(matplotlib.cm.tab20b.colors)

        default_cm = matplotlib.cm.colors.ListedColormap(default_clrs)

        # Ensure unique indices are sorted or in the desired order
        # colors_idxs = list(com_df["max_b"].unique())
        # used_colors = default_clrs[colors_idxs]

        cnorm = lambda x: x % len(default_clrs)

        used_colors = {}
        color_idxs = com_df["max_b"].unique()
        for comm in color_idxs:
            used_colors[comm] = list(default_cm(cnorm(comm)))

        pixels_x, pixels_y = size, size
        size_x, size_y = pixels_x / dpi * 2, pixels_y / dpi * 2

        # Create an empty plot with specified DPI and figure size
        fig, ax = plt.subplots(figsize=(size_x, size_y), dpi=dpi)

        # To display an empty plot with a legend, we can use a trick by plotting empty data
        for comm, color in used_colors.items():
            ax.plot([], [], color=color, label=f"Com {comm}")

        # Add legend with horizontal orientation
        ax.legend(loc='center')

        # Turn off axis
        ax.axis("off")

        return fig, ax, used_colors

    @staticmethod
    def save_legend(g: gt.Graph, path: str, name: str, size=3000, dpi=400):
        com_colors = {}
        for node in g.get_vertices():
            max_b = g.vp["max_b"][node]
            color = g.vp["fill_color"][node]
            com_colors[max_b] = color

        com_colors = dict(sorted(com_colors.items()))

        pixels_x, pixels_y = size, size
        size_x, size_y = pixels_x / dpi * 2, pixels_y / dpi * 2

        # Create an empty plot with specified DPI and figure size
        fig, ax = plt.subplots(figsize=(size_x, size_y), dpi=dpi)

        # To display an empty plot with a legend, we can use a trick by plotting empty data
        for comm, color in com_colors.items():
            ax.plot([], [], color=color, label=f"Com {comm}")

        # Add legend with horizontal orientation
        ax.legend(loc="center")

        # Turn off axis
        ax.axis("off")

        plt.savefig(f"{path}/{name}_legend.pdf")
        return com_colors

    @staticmethod
    # NOTE: this is also in GraphHelper but can't use that class as it will run in circular dependency
    def color_scale(color_scale='Sunset', num_points=100):

        import re  # Regular expression library for extracting RGB values

        # Assuming colors is a list of strings like 'rgb(75, 41, 145)'
        colors = px.colors.sample_colorscale(color_scale, samplepoints=num_points, colortype="rgb")

        # Function to convert an RGB string 'rgb(x, y, z)' to Hex
        def rgb_string_to_hex(rgb_string):
            # Extract the numerical values using regular expression
            numbers = re.findall(r"\d+", rgb_string)
            # Convert the string numbers to integers
            rgb = [int(num) for num in numbers]
            # Convert to hexadecimal
            return "#{0:02x}{1:02x}{2:02x}".format(rgb[0], rgb[1], rgb[2])

        # Convert each RGB string to Hex
        hex_colors = [rgb_string_to_hex(color) for color in colors]

        return hex_colors



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
@File    :   sankey.py
@Time    :   2022/07/04 18:10:23
@Author  :   Vlad Ungureanu
@Version :   1.0
@Contact :   vlad.ungureanu@york.ac.uk
@Desc    :   None
@Details :   This files computes and plots (w/ plotly) Sankey plots. It can received either a DataFrame or a path to .tsv containing a DataFrame
"""


import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

# load data, aggregate by subtypes across columns and create counts columnn
base_path = "/Users/vlad/Documents/Code/York/BU_clustering/results/"


def genSankey(df, cat_cols=[], value_cols="", title="Sankey Diagram", labels=None):
    """Generate the Sankey plot for a givent DataFrame

    Args:
        df (DataFrame): The DataFramed passed from which the plot is generated
        cat_cols (list, optional): The Sankey categories. Defaults to [].
        value_cols (str, optional): Index of the dataframe. Defaults to ''.
        title (str, optional): Title of the plit. Defaults to 'Sankey Diagram'.
        labels (dict, optional): Labels to be matched for each of the groups. Defaults to None.

    Returns:
        dict: Plotly figure
    """
    colorPalette = px.colors.qualitative.Vivid + px.colors.qualitative.Prism

    labelList = []
    displayLabels = []
    colorNumList = []
    for catCol in cat_cols:
        labelListTemp = list(set(df[catCol].values))
        colorNumList.append(len(labelListTemp))
        labelList = labelList + labelListTemp
        oldLabels = [labels.get(key) for key in labelListTemp]
        displayLabels = displayLabels + oldLabels

    # remove duplicates from labelList
    labelList = list(dict.fromkeys(labelList))

    # define colors based on number of levels
    colorList = []
    for idx, colorNum in enumerate(colorNumList):
        colorList = colorList + [colorPalette[idx]] * colorNum

    # transform df into a source-target pair
    for i in range(len(cat_cols) - 1):
        if i == 0:
            sourceTargetDf = df[[cat_cols[i], cat_cols[i + 1], value_cols]]
            sourceTargetDf.columns = ["source", "target", "count"]
        else:
            tempDf = df[[cat_cols[i], cat_cols[i + 1], value_cols]]
            tempDf.columns = ["source", "target", "count"]
            sourceTargetDf = pd.concat([sourceTargetDf, tempDf])
        sourceTargetDf = sourceTargetDf.groupby(["source", "target"]).agg({"count": "sum"}).reset_index()

    # add index for source-target pair
    sourceTargetDf["sourceID"] = sourceTargetDf["source"].apply(lambda x: labelList.index(x))
    sourceTargetDf["targetID"] = sourceTargetDf["target"].apply(lambda x: labelList.index(x))

    if labels == None:
        labels = displayLabels
    else:
        labels = list(labels.keys())

    # creating the sankey diagram
    data = dict(
        arrangement="perpendicular",
        type="sankey",
        node=dict(pad=15, thickness=20, line=dict(color="black", width=0.5), label=displayLabels, color=colorList),
        link=dict(source=sourceTargetDf["sourceID"], target=sourceTargetDf["targetID"], value=sourceTargetDf["count"], color="#bdbdbd"),
    )

    layout = dict(
        title=title,
        font=dict(size=16, color="#003366"),
        height=600,
        # paper_bgcolor="rgba(0,0,0,0)",
    )

    fig = dict(data=[data], layout=layout)
    return fig, (sourceTargetDf, displayLabels)


#### More descriptive functions


# Assuming figs is a list of Plotly figures and titles is a list of their titles
def prep_sankey_description(sky_meta: tuple, input_df: pd.DataFrame, sel_cols: list):
    source_target_df, labels = sky_meta[0].copy(deep=True), sky_meta[1]

    # below are the mapped labels of the groups used in Sankey (k) and the unique number given by the algorithm
    mapped_values = {k: v for k, v in enumerate(labels)}
    source_target_df["SourceLabel"] = source_target_df["sourceID"].map(mapped_values)
    source_target_df["TargetLabel"] = source_target_df["targetID"].map(mapped_values)

    # Mapp back the classifier label
    classifier_map = []
    offset = 0
    for col in sel_cols:
        for i in range(0, len(input_df[col].unique())):
            classifier_map.append((i + offset, col))
        offset += i + 1

    classifier_map = dict(classifier_map)
    # Add the information for each classifier and the size
    for source in source_target_df["source"].unique():
        source_target_df.loc[source_target_df["source"] == source, "Classifier"] = classifier_map[source]
        source_label = source_target_df.loc[source_target_df["source"] == source]["SourceLabel"].values[0]
        group_size = input_df.loc[input_df[classifier_map[source]] == source_label].shape[0]

        source_target_df.loc[source_target_df["source"] == source, "source_size"] = int(group_size)

    source_target_df["prct_chg"] = round(source_target_df["count"] / source_target_df["source_size"], 4)
    source_target_df["SourceLabel"] = source_target_df["SourceLabel"].astype(str) + "(" + source_target_df["source_size"].astype(int).astype(str) + ")"

    source_target_df.rename(columns={"count": "#Samples"}, inplace=True)
    return source_target_df


def plot_sankey_description(config: dict, sky_stats: pd.DataFrame):
    figs, titles = [], []

    max_clusters = max([len(sky_stats[sky_stats["Classifier"] == elem]["SourceLabel"].unique()) for elem in sky_stats["Classifier"].unique()])

    titles = [["" for _ in range(max_clusters)] for _ in sky_stats["Classifier"].unique()]
    for idx, classifier in enumerate(sky_stats["Classifier"].unique()):
        sel_df = sky_stats.loc[sky_stats["Classifier"] == classifier]
        sel_df["TargetLabel"] = sel_df["TargetLabel"].astype(str)
        fig = px.bar(sel_df, x="TargetLabel", y="#Samples", facet_col="SourceLabel", text_auto=True)
        figs.append(fig)
        # titles.append(f"{classifier}")
        for label_idx, label in enumerate(sel_df["SourceLabel"].unique()):
            titles[idx][label_idx] = label

    titles = np.array(titles).flatten()

    if config is None:
        config = {
            "shared_x": False,
            "shared_y": False,
            "h_spacing": 0.02,
            "v_spacing": 0.15,
            "main_title": "Subplots",
            "height": 700,
            "width": None,
            "y_title": "Y-axis",
            "x_title": "X-axis",
            "specs": None,
        }

    # Determine the total number of rows and columns
    num_rows = len(figs)
    num_cols = max(len(fig.data) for fig in figs)  # Assuming each fig has the same number of columns

    # Create a subplot grid
    subplot_titles = [""] * (num_rows * num_cols)

    # Set the title for the first subplot in each row
    for i in range(num_rows):
        subplot_titles[i * num_cols + round(num_cols / 2)] = titles[i]

    # Create a subplot grid with titles
    subplot = make_subplots(
        rows=num_rows,
        cols=num_cols,
        subplot_titles=titles,
        shared_xaxes=config["shared_x"],
        shared_yaxes=config["shared_y"],
        horizontal_spacing=config["h_spacing"],
        vertical_spacing=config["v_spacing"],
    )

    # Add each trace to the correct position
    for i, single_fig in enumerate(figs):
        for j, trace in enumerate(single_fig.data):
            subplot.add_trace(trace, row=i + 1, col=j + 1)

    # Update layout if necessary (e.g., titles, axes labels)
    layout = go.Layout(title_text=config["main_title"])
    subplot = subplot.update_layout(layout, height=config["height"], width=config["width"], showlegend=False)
    # subplot = subplot.update_xaxes(title_text=config["x_title"])
    # subplot = subplot.update_yaxes(title_text=config["y_title"])
    return subplot


def main(df=None, filename="TCGA_classifier_labels.tsv", title="TCGA MIBC mRNA subtypes", drop_cols=None, reorder_cols=None, retMeta=False):
    """Function which process the dataframe and get it ready to generate a Saneku plot

    Args:
        df (DataFrame, optional): The DataFrame from where the Saneky is generated. Defaults to None.
        filename (str, optional): The filePath to a .tsv file from which the Saneky is generated.. Defaults to "TCGA_classifier_labels.tsv".
        title (str, optional): Title of the plot. Defaults to "TCGA MIBC mRNA subtypes".
        drop_cols (_type_, optional): Remove the specified groups from the Sankey. Defaults to None.
        reorder_cols (_type_, optional): Re-arrange the groups in the Sankey. Defaults to None.
    """

    if df is None:
        classDF = pd.read_csv(base_path + filename, sep="\t", index_col="Sample")
    else:
        classDF = df

    if drop_cols != None:
        classDF.drop(columns=drop_cols, axis=1, inplace=True)

    if reorder_cols != None:
        classDF = classDF[reorder_cols].copy(deep=True)

    oricols = list(classDF.columns)
    uniq_list = []
    remap_labels = {}
    # keeps track of the number of groups in the previous columns
    init = 0
    # go through each column, and the values in that column (2nd loop) and make sure that each label has an unique value
    for col in list(classDF.columns):
        col_uniq = sorted(list(classDF[col].unique()))
        uniq_list += col_uniq
        col_rep = list(classDF[col])
        idx = 0
        for label in col_uniq:
            new_idx = init + idx
            # There are some cases where a new value for the label might be with something that is already in the column; This will break the code where we do re-labeling.
            if new_idx in col_uniq and label != new_idx:
                new_idx += 1
                idx += 1

            remap_labels[new_idx] = label
            col_rep = [new_idx if x == label else x for x in col_rep]
            idx += 1

        classDF[f"{col}_idx"] = col_rep
        init += idx

    # remove the originals and groupby rows to generate counts
    classDFidx = classDF.drop(oricols, axis=1)
    classDFidxagg = classDFidx.groupby(list(classDFidx.columns)).size().reset_index(name="counts")

    # if show_labels:
    fig_dict, meta = genSankey(classDFidxagg, cat_cols=list(classDFidx.columns), value_cols="counts", title=title, labels=remap_labels)
    fig = go.Figure(fig_dict)

    # add it column names
    for x_coordinate, col_name in enumerate(classDFidx.columns):
        col_name = "_".join(col_name.split("_")[:-1])
        fig.add_annotation(
            x=x_coordinate,
            y=1.07,
            xref="x",
            yref="paper",
            text=col_name,
            showarrow=False,
            font=dict(size=18, color="#003366"),
            align="left",
        )

    fig = fig.update_layout(
        xaxis={
            "showgrid": False,  # thin lines in the background
            "zeroline": False,  # thick line at x=0
            "visible": False,  # numbers below
        },
        yaxis={
            "showgrid": False,  # thin lines in the background
            "zeroline": False,  # thick line at x=0
            "visible": False,  # numbers below
        },
        plot_bgcolor="rgba(0,0,0,0)",
        template="ggplot2",
        font=dict(size=16),
    )

    if retMeta:
        return meta, fig
    else:
        
        fig.show()

    # pio.write_image(fig, base_path + "TCGA_classifications_no-ori.pdf"
